///////////////////////////////////////////////////////////////////////
// QuickTime plugin for VirtualDub 1.8.8+
// Copyright (C) 2007-2012 fccHandler
// Copyright (C) 1998-2012 Avery Lee
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
///////////////////////////////////////////////////////////////////////

// Parser for the H.264 Sequence Parameter Set.

// Define NULL to avoid bringing in windows.h
#define  NULL 0

#include "AVCSequence.h"

void AVCSequence::Reset()
{
	pos                            = 0;
	curbyte                        = 0;
	curbit                         = 0;
	error                          = 0;

	profile_idc                    = 0;
	separate_colour_plane_flag     = 0;
	frame_mbs_only_flag            = 1;

	ChromaArrayType                = CHROMA420;
	SubWidthC                      = 2;
	SubHeightC                     = 2;
	CropUnitX                      = 2;
	CropUnitY                      = 2;

	chroma_format_idc              = CHROMA420;
	pic_width_in_mbs_minus1        = 0;
	pic_height_in_map_units_minus1 = 0;

	frame_crop_left_offset         = 0;
	frame_crop_right_offset        = 0;
	frame_crop_top_offset          = 0;
	frame_crop_bottom_offset       = 0;
}


unsigned char AVCSequence::next_bit()
{
	unsigned char ret = 0;

	if (curbit == 0)
	{
		if (bits == NULL) goto Abort;
		if (pos  >= len ) goto Abort;

		curbyte = bits[pos++];
		curbit  = 128;
	}

	if (curbyte & curbit) ++ret;
	curbit >>= 1;
	return ret;

Abort:
	error = 1;
	return 1;
}


unsigned int AVCSequence::get_bits(int n)
{
	unsigned int ret = 0;
	while (n > 0)
	{
		ret = (ret * 2) + next_bit();
		--n;
	}
	return ret;
}


unsigned int AVCSequence::get_ue_golomb()
{
	// Return the next unsigned Exp-Golomb code.
	// (cf. ITU-T Rec. H.264, section 9.1)

	unsigned int b;
	unsigned int codeNum;
	int          leadingZeroBits = -1;

	for (b = 0; !b; ++leadingZeroBits)
	{
		b = next_bit();
	}

	codeNum = ((1U << leadingZeroBits) - 1) + get_bits(leadingZeroBits);
	return codeNum;
}


int AVCSequence::get_se_golomb()
{
	// Return the next signed Exp-Golomb code.
	// (cf. ITU-T Rec. H.264, section 9.1.1)

	int codeNum = (int)get_ue_golomb();
	return ((codeNum + 1) >> 1) * ((codeNum & 1)? 1: -1);
}

void AVCSequence::scaling_list(int sizeOfScalingList)
{
	// Parse a scaling list (discarded).
	// (cf. ITU-T Rec. H.264, section 7.3.2.1.1.1)

	int delta_scale;
	int j;

	unsigned int scalingListj;
	unsigned int lastScale = 8;
	unsigned int nextScale = 8;

	for (j = 0; j < sizeOfScalingList; ++j)
	{
		if (nextScale != 0)
		{
			delta_scale = get_se_golomb();
			nextScale = (lastScale + delta_scale + 256) % 256;
			//useDefaultScalingMatrixFlag[j] = (j == 0 && nextScale == 0);
		}
		scalingListj = (nextScale == 0)? lastScale: nextScale;
		lastScale = scalingListj;
	}
}


bool AVCSequence::ParseSPS(const void *data, int size)
{
	// Parse one sequence parameter set.
	// (cf. ITU-T Rec. H.264, section 7.3.2.1.1)

	unsigned int i;

	bits = (unsigned char *)data;
	len  = size;
	Reset();

	i = next_bit();             // forbidden_zero_bit
	if (i == 1) error = 1;      // (must be zero)

	i = get_bits(2);            // nal_ref_idc
	if (i == 0) error = 1;      // (must not be zero)

	i = get_bits(5);            // nal_unit_type
	if (i != 7) error = 1;      // (7 = sequence parameter set)

	profile_idc = (unsigned char)get_bits(8);

	next_bit();                 // constraint_set0_flag
	next_bit();                 // constraint_set1_flag
	next_bit();                 // constraint_set2_flag
	next_bit();                 // constraint_set3_flag
	next_bit();                 // constraint_set4_flag
	next_bit();                 // constraint_set5_flag
	get_bits(2);                // reserved_zero_2bits
	get_bits(8);                // level_idc
	get_ue_golomb();            // seq_parameter_set_id

	if (profile_idc ==  44 ||   /* CAVLC 4:4:4       */
	    profile_idc ==  83 ||   /* Scalable Baseline */
	    profile_idc ==  86 ||   /* Scalable High     */
	    profile_idc == 100 ||   /* High              */
	    profile_idc == 110 ||   /* High 10           */
	    profile_idc == 122 ||   /* High 4:2:2        */
	    profile_idc == 244 )    /* High 4:4:4        */
	{
		chroma_format_idc = get_ue_golomb();
		if (chroma_format_idc == CHROMA444)
		{
			separate_colour_plane_flag = next_bit();
		}
		get_ue_golomb();        // bit_depth_luma_minus8
		get_ue_golomb();        // bit_depth_chroma_minus8
		next_bit();             // qpprime_y_zero_transform_bypass_flag
		if (next_bit())         // seq_scaling_matrix_present_flag
		{
			for (i = 0; i < ((chroma_format_idc != CHROMA444)? 8U: 12U); ++i)
			{
				if (next_bit()) // seq_scaling_list_present_flag[i]
				{
					if (i < 6)
						scaling_list(/*ScalingList4x4[i],*/ 16 /*, UseDefaultScalingMatrix4x4Flag[i]*/);
					else
						scaling_list(/*ScalingList8x8[i - 6],*/ 64 /*, UseDefaultScalingMatrix8x8Flag[i - 6]*/);
				}
			}
		}
	}

	get_ue_golomb();            // log2_max_frame_num_minus4
	i = get_ue_golomb();        // pic_order_cnt_type
	if (i == 0)
	{
		get_ue_golomb();        // log2_max_pic_order_cnt_lsb_minus4
	}
	else if (i == 1)
	{
		next_bit();             // delta_pic_order_always_zero_flag
		get_se_golomb();        // offset_for_non_ref_pic
		get_se_golomb();        // offset_for_top_to_bottom_field
		i = get_ue_golomb();    // num_ref_frames_in_pic_order_cnt_cycle
		while (i--)
		{
			get_se_golomb();    // offset_for_ref_frame[i]
		}
	}

	get_ue_golomb();        // max_num_ref_frames
	next_bit();             // gaps_in_frame_num_value_allowed_flag

	pic_width_in_mbs_minus1         = get_ue_golomb();
	pic_height_in_map_units_minus1  = get_ue_golomb();

	frame_mbs_only_flag = next_bit();
	if (!frame_mbs_only_flag)
	{
		next_bit();         // mb_adaptive_frame_field_flag
	}

	next_bit();             // direct_8x8_inference_flag
	if (next_bit())         // frame_cropping_flag
	{
		frame_crop_left_offset   = get_ue_golomb();
		frame_crop_right_offset  = get_ue_golomb();
		frame_crop_top_offset    = get_ue_golomb();
		frame_crop_bottom_offset = get_ue_golomb();
	}

	// vui_parameters may follow

	if (separate_colour_plane_flag)
		ChromaArrayType = 0;
	else
		ChromaArrayType = chroma_format_idc;

	switch (chroma_format_idc)
	{
		case 0:
			break;
		case CHROMA420:
			SubWidthC  = 2;
			SubHeightC = 2;
			break;
		case CHROMA422:
			SubWidthC  = 2;
			SubHeightC = 1;
			break;
		case CHROMA444:
			if (!separate_colour_plane_flag)
			{
				SubWidthC  = 1;
				SubHeightC = 1;
			}
			break;
	}

	if (ChromaArrayType == 0)
	{
		CropUnitX = 1;
		CropUnitY = 2 - frame_mbs_only_flag;
	}
	else
	{
		CropUnitX = SubWidthC;
		CropUnitY = SubHeightC * (2 - frame_mbs_only_flag);
	}

	return (error == 0);
}

int AVCSequence::Width() const
{
	return ((pic_width_in_mbs_minus1 + 1) * 16)
	     - ((frame_crop_left_offset * CropUnitX)
	     + (frame_crop_right_offset * CropUnitX));
}

int AVCSequence::Height() const
{
	return ((2 - frame_mbs_only_flag)
	     * (pic_height_in_map_units_minus1 + 1) * 16)
	     - ((frame_crop_top_offset * CropUnitY)
	     + (frame_crop_bottom_offset * CropUnitY));
}

