///////////////////////////////////////////////////////////////////////
// QuickTime plugin for VirtualDub 1.8.8+
// Copyright (C) 2007-2012 fccHandler
// Copyright (C) 1998-2012 Avery Lee
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
///////////////////////////////////////////////////////////////////////

// IVDXAudioSource interface

#define WIN32_LEAN_AND_MEAN
#include <windows.h>
#include <commctrl.h>
#include <vfw.h>
#include <stdio.h>

#include "vd2\plugin\vdinputdriver.h"
#include "Unknown.h"

#include "QTTypes.h"
#include "QTTrack.h"

#include "WinFile.h"
#include "ErrorLog.h"
#include "InputOptionsQT.h"
#include "InputFileQT.h"
#include "AudioSourceQT.h"

#include <crtdbg.h>

//#define ENABLE_ADPCM_DSC

/********************************************************
 * AudioSourceQT (default implementation)
 * This class handles:
 *  '.mp3'  MPEG Layer III (CBR / VBR)
 *  'ac-3'  Dolby Digital AC-3
 *  'mp4a'  Advanced Audio Codec
 *  'ms  '  Microsoft IMA ADPCM
 *  'ms U'  MPEG Layer III (CBR only)
 *  'samr'  AMR (narrowband)
 ********************************************************/

AudioSourceQT::AudioSourceQT(InputFileQT *const pp, QTTrack *const qtt)
	: parentPtr(pp)
	, audio(qtt)
{
// Note: parent won't create us unless audio track is valid
	parentPtr->AddRef();
	errmode = kErrorModeDecodeAnyway;
}

AudioSourceQT::~AudioSourceQT()
{
	parentPtr->Release();
}

int AudioSourceQT::AddRef()
{
	return vdxunknown<IVDXStreamSource>::AddRef();
}

int AudioSourceQT::Release()
{
	return vdxunknown<IVDXStreamSource>::Release();
}

void *VDXAPIENTRY AudioSourceQT::AsInterface(uint32 iid)
{
	if (iid == IVDXAudioSource::kIID)
		return static_cast<IVDXAudioSource *>(this);

	if (iid == IVDXStreamSourceV3::kIID)
		return static_cast<IVDXStreamSourceV3 *>(this);

	return vdxunknown<IVDXStreamSource>::AsInterface(iid);
}

void VDXAPIENTRY AudioSourceQT::GetStreamSourceInfo(VDXStreamSourceInfo& srcInfo)
{
	srcInfo.mSampleRate.mNumerator         = audio->dwRate;
	srcInfo.mSampleRate.mDenominator       = audio->dwScale;
	srcInfo.mSampleCount                   = audio->perceived_samples;
	srcInfo.mPixelAspectRatio.mNumerator   = 1;
	srcInfo.mPixelAspectRatio.mDenominator = 1;
}

void VDXAPIENTRY AudioSourceQT::GetStreamSourceInfoV3(VDXStreamSourceInfoV3& info)
{
	GetStreamSourceInfo(info.mInfo);

	if (audio->dwFlags & QTT_HAS_VSB)
		info.mFlags = VDXStreamSourceInfoV3::kFlagVariableSizeSamples;
	else
		info.mFlags = 0;

	info.mfccHandler = 0;
}

bool AudioSourceQT::Read(sint64 lStart64, uint32 lCount, void *lpBuffer, uint32 cbBuffer, uint32 *lBytesRead, uint32 *lSamplesRead)
{
	const QTSampleInfo *si;
	unsigned char *dst;

	unsigned long dwStart;
	unsigned long actual;
	unsigned long i;

	// ADTS header stuff
	unsigned long dwAdtsSize;
	unsigned int  profile, chcode, srcode;

	dwStart = (DWORD)lStart64;

_RPT3(_CRT_WARN, "dwStart = %lu, lCount = %lu, cbBuffer = %lu\n", dwStart, lCount, cbBuffer);

	// Start out of range?
	if (dwStart >= audio->num_samples)
	{
_RPT1(_CRT_WARN, "dwStart >= num_samples (%lu)\n", audio->num_samples);

		if (lBytesRead) *lBytesRead = 0;
		if (lSamplesRead) *lSamplesRead = 0;
		return true;
	}

	// Add ADTS headers to AAC?
	dwAdtsSize = (audio->dwFlags & QTT_ADD_ADTS)? 7: 0;

	// Sample count out of range?
	actual = audio->num_samples - dwStart;
	if (lCount > actual) lCount = actual;

	si = &(audio->sample_list[dwStart]);

	// Buffer size query?
	if (lpBuffer == NULL)
	{
		actual = si->size + dwAdtsSize;

		for (i = 1; i < lCount; ++i)
		{
			++si;
			actual += si->size + dwAdtsSize;
		}

		if (dwStart == 0 && (audio->dwFlags & QTT_ADD_AMR))
		{
			// Add AMR header
			if (audio->dwFormat == A_SAMR)
				actual += 6;
			else if (audio->dwFormat == A_SAWB)
				actual += 9;
		}

		if (actual < audio->dwMaxSample) actual = audio->dwMaxSample;

_RPT2(_CRT_WARN, "lpBuffer is NULL, returning lBytesRead = %lu, lSamplesRead = %lu\n", actual, i);

		if (lBytesRead) *lBytesRead = actual;
		if (lSamplesRead) *lSamplesRead = i;
		return true;
	}

	// Buffer too small to hold a sample?
	if (cbBuffer < audio->dwMaxSample)
	{
		if (lBytesRead) *lBytesRead = audio->dwMaxSample;
		if (lSamplesRead) *lSamplesRead = 1;
		return false;
	}

	if (dwAdtsSize != 0)
	{
		// Gather information for ADTS header	

		// AudioSpecificConfig:
		//   LSB      MSB
		// xxxxx... ........  object type (00010 = AAC LC)
		// .....xxx x.......  sampling rate
		// ........ .xxxx...  channel configuration
		// ........ .....x..  frame length flag (0 = 1024, 1 = 960)
		// ........ ......x.  depends on core coder (0)
		// ........ .......x  extension flag (0)

		profile = *((WORD *)(audio->pFormat + sizeof(WAVEFORMATEX)));
		profile = ((profile & 0xFF) << 8) | (profile >> 8);

		chcode  = (profile >> 3) & 7;
		srcode  = (profile >> 7) & 15;
		profile = ((profile >> 11) - 1) & 3;
	}

	// Begin reading whole samples
	dst = (unsigned char *)lpBuffer;

	if (dwStart == 0 && (audio->dwFlags & QTT_ADD_AMR))
	{
		// Add AMR header?
		if (audio->dwFormat == A_SAMR)
		{
			memcpy(dst, "#!AMR\n", 6);
			dst += 6;
			cbBuffer -= 6;
		}
		else if (audio->dwFormat == A_SAWB)
		{
			memcpy(dst, "#!AMR-WB\n", 9);
			dst += 9;
			cbBuffer -= 9;
		}
	}

	for (i = 0; i < lCount; ++i)
	{
		if (cbBuffer < (si->size + dwAdtsSize)) break;

		if (dwAdtsSize != 0)
		{
			// Build ADTS header:
			// 00: 11111111   syncword
			// 01: 1111....   syncword (continued)
			//     ....0...   ID (1 = MPEG-2, 0 = MPEG-4)
			//     .....00.   layer
			//     .......1   protection absent
			// 02: xx......   profile
			//     ..xxxx..   sampling frequency index
			//     ......0.   private bit
			//     .......x   channel configuration
			// 03: xx......   channel configuration (continued)
			//     ..0.....   original / copy
			//     ...0....   home
			//     ....0...   copyright identification bit
			//     .....0..   copyright identification start
			//     ......xx   frame length
			// 04: xxxxxxxx   frame length (continued)
			// 05: xxx.....   frame length (continued)
			//     ...11111   buffer fullness
			// 06: 111111..   buffer fullness (continued)
			//     ......00   number of raw data blocks in frame

			// Nero Wave Editor creates files with ID = 1 (???)

			unsigned int frlen = (7 + si->size) & 0x1FFF;
			
			dst[0] = 0xFF;
			dst[1] = 0xF1;
			dst[2] = (unsigned char)((profile << 6) | (srcode << 2) | (chcode >> 2));
			dst[3] = (unsigned char)(((chcode & 3) << 6) | ((frlen >> 11) & 3));
			dst[4] = (unsigned char)((frlen >> 3) & 0xFF);
			dst[5] = (unsigned char)(((frlen & 7) << 5) | 0x1F);
			dst[6] = 0xFC;
		}

		actual = parentPtr->ReadWholeSample(dst + dwAdtsSize, *si);
		if (actual != si->size) break;	// end of file?

		actual   += dwAdtsSize;
		dst      += actual;
		cbBuffer -= actual;

		++si;
	}

	actual = (unsigned long)(dst - (unsigned char *)lpBuffer);

_RPT2(_CRT_WARN, "Returning lBytesRead = %lu, lSamplesRead = %lu\n", actual, i);

	if (lBytesRead) *lBytesRead = actual;
	if (lSamplesRead) *lSamplesRead = i;

	return true;
}

const void *AudioSourceQT::GetDirectFormat()
{
	return audio->pFormat;
}

int AudioSourceQT::GetDirectFormatLen()
{
	return audio->lFormatLen;
}

IVDXStreamSource::ErrorMode AudioSourceQT::GetDecodeErrorMode()
{
	return errmode;
}

void AudioSourceQT::SetDecodeErrorMode(IVDXStreamSource::ErrorMode mode)
{
	errmode = mode;
}

bool AudioSourceQT::IsDecodeErrorModeSupported(IVDXStreamSource::ErrorMode mode)
{
	return true;
}

bool AudioSourceQT::IsVBR()
{
	return false;
}

sint64 AudioSourceQT::TimeToPositionVBR(sint64 us)
{
	/*
	const WAVEFORMATEX *wfex = (WAVEFORMATEX *)audio->pFormat;

	// (sample_rate * seconds) / samples_per_frame

	return (sint64)(
		(us * (double)wfex->nSamplesPerSec)
		/ (audio->dwSamplesPerPacket * 1000000.0) + 0.5
	);
	*/
	return 0;
}

sint64 AudioSourceQT::PositionToTimeVBR(sint64 samples)
{
	/*
	const WAVEFORMATEX *wfex = (WAVEFORMATEX *)audio->pFormat;

	// (samples * samples_per_frame) / sample_rate

	return (sint64)(
		(1000000.0 * samples * audio->dwSamplesPerPacket)
		/ (double)wfex->nSamplesPerSec + 0.5
	);
	*/
	return 0;
}

void AudioSourceQT::GetAudioSourceInfo(VDXAudioSourceInfo& info)
{
	info.mFlags = 0;
}


/********************************************************
 * AudioSource_PCM
 * This class handles:
 *  'alaw'  8-bit aLaw 2:1
 *  'fl32'  32-bit float
 *  'fl64'  64-bit float
 *  'in24'  24-bit PCM
 *  'in32'  32-bit PCM
 *  'raw '  8-bit PCM
 *  'sowt'  16-bit PCM (little-endian)
 *  'twos'  16-bit PCM (big-endian)
 *  'ulaw'  8-bit uLaw 2:1
 ********************************************************/

#if defined _MSC_VER && _MSC_VER >= 1400
	#pragma intrinsic(_byteswap_ushort)
	#pragma intrinsic(_byteswap_ulong)
#else
	#define _byteswap_ushort(x) (unsigned short)(       \
	      ((unsigned short)(x) >> 8)                    \
	    | ((unsigned short)(x) << 8))

	#define _byteswap_ulong(x) (unsigned long)(         \
	      ( (unsigned long)(x) >> 24)                   \
	    | (((unsigned long)(x) >> 8 ) & 0x0000FF00UL)   \
	    | (((unsigned long)(x) << 8 ) & 0x00FF0000UL)   \
	    | ( (unsigned long)(x) << 24))
#endif

void AudioSource_PCM::PCM_word_swap(void *buffer, int bytes)
{
	if (buffer != NULL)
	{
		unsigned short *src = (unsigned short *)buffer;

		while (bytes >= 2)
		{
			unsigned short sample = *src;
			*src++ = _byteswap_ushort(sample);
			bytes -= 2;
		}
	}
}

void AudioSource_PCM::PCM_fword_swap(void *buffer, int bytes)
{
	if (buffer != NULL)
	{
		unsigned char *src = (unsigned char *)buffer;

		while (bytes >= 3)
		{
			unsigned char samp1 = src[0];
			unsigned char samp2 = src[2];
			*src++ = samp2;
			++src;
			*src++ = samp1;
			bytes -= 3;
		}
	}
}

void AudioSource_PCM::PCM_dword_swap(void *buffer, int bytes)
{
	if (buffer != NULL)
	{
		unsigned long *src = (unsigned long *)buffer;

		while (bytes >= 4)
		{
			unsigned long sample = *src;
			*src++ = _byteswap_ulong(sample);
			bytes -= 4;
		}
	}
}

void AudioSource_PCM::PCM_qword_swap(void *buffer, int bytes)
{
	if (buffer != NULL)
	{
		unsigned long *src = (unsigned long *)buffer;

		while (bytes >= 8)
		{
			unsigned long samp1 = src[0];
			unsigned long samp2 = src[1];
			*src++ = _byteswap_ulong(samp2);
			*src++ = _byteswap_ulong(samp1);
			bytes -= 8;
		}
	}
}

bool AudioSource_PCM::Read(sint64 lStart64, uint32 lCount, void *lpBuffer, uint32 cbBuffer, uint32 *lBytesRead, uint32 *lSamplesRead)
{
	const WAVEFORMATEX *wfex = (WAVEFORMATEX *)audio->pFormat;
	const unsigned int ba = wfex->nBlockAlign;

	unsigned int actual;

	// too small to hold a sample?
	if (lpBuffer == NULL || cbBuffer < ba)
	{
		actual = lCount * ba;
		if (actual < ba) actual = ba;

		if (lBytesRead) *lBytesRead = actual;
		if (lSamplesRead) *lSamplesRead = actual / ba;

		return (lpBuffer == NULL);
	}

	actual = parentPtr->ReadRawStream(lpBuffer, lStart64 * ba, lCount * ba, *audio);

	if (wfex->wBitsPerSample == 8 && (audio->dwFormat == A_SOWT || audio->dwFormat == A_TWOS))
	{
		// convert signed 8-bit to unsigned 8-bit
		unsigned int i;
		unsigned char *src = (unsigned char *)lpBuffer;

		for (i = 0; i < actual; ++i) src[i] += 128;
	}
	else if (audio->iEndianness == 0)
	{
		// swap big-endian to little-endian
		switch (wfex->wBitsPerSample) {
			case 16:
				PCM_word_swap(lpBuffer, actual);
				break;
			case 24:
				PCM_fword_swap(lpBuffer, actual);
				break;
			case 32:
				PCM_dword_swap(lpBuffer, actual);
				break;
			case 64:
				PCM_qword_swap(lpBuffer, actual);
				break;
		}
	}

	if (lBytesRead) *lBytesRead = actual;
	if (lSamplesRead) *lSamplesRead = actual / ba;

	return true;
}


/********************************************************
 * AudioSource_ima4
 * This class handles:
 *  'ima4'  QuickTime IMA ADPCM
 ********************************************************/

AudioSource_ima4::AudioSource_ima4(InputFileQT *const pp, QTTrack *const qtt)
	: AudioSourceQT(pp, qtt)
	, inp_blksize((2 + (qtt->dwSamplesPerPacket >> 1)) * ((WAVEFORMATEX *)qtt->pFormat)->nChannels)
#ifdef ENABLE_ADPCM_DSC
	, out_blksize((2 + (qtt->dwSamplesPerPacket >> 1)) * ((WAVEFORMATEX *)qtt->pFormat)->nChannels)
#else
	, out_blksize(qtt->dwSamplesPerPacket * ((WAVEFORMATEX *)qtt->pFormat)->nBlockAlign)
#endif
{
	int boundary = (inp_blksize + 15) & -16;		

	adpcm_out = NULL;
	adpcm_buf = new unsigned char[boundary + out_blksize];

	if (adpcm_buf != NULL) adpcm_out = adpcm_buf + boundary;
}

AudioSource_ima4::~AudioSource_ima4()
{
	delete[] adpcm_buf;
}

void AudioSource_ima4::adpcm_decode()
{
	// Decode samples from adpcm_buf to adpcm_out.

	static const signed char index_table[8] = {
		-1, -1, -1, -1, 2, 4, 6, 8
	};

	static const unsigned short step_table[89] = {
		7, 8, 9, 10, 11, 12, 13, 14, 16, 17,
		19, 21, 23, 25, 28, 31, 34, 37, 41, 45,
		50, 55, 60, 66, 73, 80, 88, 97, 107, 118,
		130, 143, 157, 173, 190, 209, 230, 253, 279, 307,
		337, 371, 408, 449, 494, 544, 598, 658, 724, 796,
		876, 963, 1060, 1166, 1282, 1411, 1552, 1707, 1878, 2066,
		2272, 2499, 2749, 3024, 3327, 3660, 4026, 4428, 4871, 5358,
		5894, 6484, 7132, 7845, 8630, 9493, 10442, 11487, 12635, 13899,
		15289, 16818, 18500, 20350, 22385, 24623, 27086, 29794, 32767
	};

	const WAVEFORMATEX *wfex = (WAVEFORMATEX *)audio->pFormat;
	const unsigned char *src = adpcm_buf;

	unsigned int count;
	unsigned int chan;

#ifdef ENABLE_ADPCM_DSC
	memcpy(adpcm_out, src, inp_blksize);
	return;
#endif

	for (chan = 0; chan < wfex->nChannels; ++chan)
	{
		unsigned char nibble;
		int predictor;
		int step_index;

		short *dst = (short *)adpcm_out;
		dst += chan;

		step_index = *src++ << 8;
		step_index |= *src++;

		predictor = step_index & 0xFF80;
		step_index &= 0x7F;

		if (predictor & 0x8000) predictor |= 0xFFFF0000;
		if (step_index > 88) step_index = 88;

	    for (count = 0; count < audio->dwSamplesPerPacket; ++count)
		{
			int diff;
			int step = step_table[step_index];

			if (count & 1)
				nibble >>= 4;
			else
				nibble = *src++;

			step_index += index_table[nibble & 7];

			if (step_index < 0)
				step_index = 0;
			else if (step_index > 88)
				step_index = 88;

			// There are at least two common ways to
			// perform the "ideal" conversion we want:
			// diff = ((nibble & 7) + 0.5) * step / 4.0

			// MPlayer (and presumably ffdshow) use the following:
			// diff = ((2 * (nibble & 7) + 1) * step) >> 3;
			
			// Microsoft's ADPCM codec uses the following:
			diff = step >> 3;
			if (nibble & 4) diff += step;
			if (nibble & 2) diff += (step >> 1);
			if (nibble & 1) diff += (step >> 2);

			// Notice there are minor differences in the results
			// of the two algorithms! Therefore I have decided
			// to stick with the "classic" Microsoft algorithm.

			if (nibble & 8)
				predictor -= diff;
			else
				predictor += diff;

			if (predictor < -32768)
				predictor = -32768;
			else if (predictor > 32767)
				predictor = 32767;

            *dst = (short)predictor;
            dst += wfex->nChannels;
        }
    }
}


bool AudioSource_ima4::Read(sint64 lStart64, uint32 lCount, void *lpBuffer, uint32 cbBuffer, uint32 *lBytesRead, uint32 *lSamplesRead)
{
	const WAVEFORMATEX *wfex = (WAVEFORMATEX *)audio->pFormat;
	const unsigned int ba = wfex->nBlockAlign;

	unsigned int actual;
	unsigned int remains;
	int offset;
	
	if (adpcm_buf == NULL)
	{
		if (lBytesRead) *lBytesRead = 0;
		if (lSamplesRead) *lSamplesRead = 0;
		return true;
	}

	// buffer size query or buffer too small?
	if (lpBuffer == NULL || cbBuffer < ba)
	{
		actual = lCount * ba;
		if (actual < ba) actual = ba;

		if (lBytesRead) *lBytesRead = actual;
		if (lSamplesRead) *lSamplesRead = actual / ba;

		return (lpBuffer == NULL);
	}

	// Absolute byte position, bytes to read
	lStart64 *= ba;
	lCount   *= ba;

	offset   = (int)(lStart64 % out_blksize);
	lStart64 = (lStart64 / out_blksize) * inp_blksize;
	actual   = 0;

	while (lCount != 0)
	{
		remains = parentPtr->ReadRawStream(adpcm_buf, lStart64, inp_blksize, *audio);
		if (remains != inp_blksize) break;

		adpcm_decode();

		remains = out_blksize - offset;
		if (remains > lCount) remains = lCount;

		memcpy(lpBuffer, adpcm_out + offset, remains);

		lCount -= remains;
		actual += remains;

		lpBuffer = (char *)lpBuffer + remains;
		lStart64 += inp_blksize;
		offset = 0;
	}

	if (lBytesRead) *lBytesRead = actual;
	if (lSamplesRead) *lSamplesRead = actual / ba;

	return true;
}


