///////////////////////////////////////////////////////////////////////
// QuickTime plugin for VirtualDub 1.8.8+
// Copyright (C) 2007-2012 fccHandler
// Copyright (C) 1998-2012 Avery Lee
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
///////////////////////////////////////////////////////////////////////

// IVDXInputFile implementation


#if defined(_MSC_VER) && (_MSC_VER >= 1400)
	#define _CRT_SECURE_NO_WARNINGS
#endif

#define WIN32_LEAN_AND_MEAN
#include <windows.h>
#include <commctrl.h>
#include <vfw.h>
#include <stdio.h>

#include "vd2\plugin\vdinputdriver.h"
#include "Unknown.h"

#include "QTTypes.h"
#include "QTTrack.h"

#include "WinFile.h"
#include "ErrorLog.h"
#include "InputOptionsQT.h"
#include "InputFileQT.h"
#include "VideoSourceQT.h"
#include "AudioSourceQT.h"
#include "AVCSequence.h"
#include "Resource.h"

#include <crtdbg.h>


// zlib
extern "C" int uncompress (unsigned char *dest, unsigned long *destLen,
	const unsigned char *source, unsigned long sourceLen);


HMODULE g_hModule;

BOOL WINAPI DllMain(HINSTANCE hinstDLL, DWORD fdwReason, LPVOID lpvReserved)
{
	if (fdwReason == DLL_PROCESS_ATTACH) g_hModule = (HMODULE)hinstDLL;
	return TRUE;
}


// Required:
//	trak
//		tkhd
//		mdia
//			mdhd


///////////////////////////////////////////////////////////////////////
//
//						InputFileQT
//
///////////////////////////////////////////////////////////////////////


InputFileQT::InputFileQT(const VDXInputDriverContext& context)
	: mContext      (context)
	, major_brand   (0)
	, minor_version (0)
	, compat_brand  (0)
	, pOptions      (NULL)
	, moov_bytes    (NULL)
	, moov_pos      (0)
	, moov_size     (0)
	, vCodec        (NULL)
	, num_tracks    (0)
{
	for (int i = 0; i < MAX_TRACKS; ++i)
	{
		Track[i] = NULL;
	}
}


InputFileQT::~InputFileQT()
{
	for (int i = num_tracks; i > 0; )
	{
		delete Track[--i];
	}

	delete[] moov_bytes;

	if (pOptions != NULL) pOptions->Release();
}


int InputFileQT::moov_read(void *dst, int bytes)
{
	if (dst != NULL
		&& bytes > 0
		&& moov_bytes != NULL
		&& moov_pos < moov_size)
	{
		const unsigned int remains = moov_size - moov_pos;
		if ((unsigned int)bytes > remains) bytes = remains;

		memcpy(dst, moov_bytes + moov_pos, bytes);
		moov_pos += bytes;

		return bytes;
	}
	return 0;
}


bool InputFileQT::NextAtom(QTAtom& atom)
{
	// Return the next atom in the file

	bool ret = false;
	__int64 sz = 0;

	atom.offset = mFile.Pos();
	if (mFile.Read(&sz, 4) != 4) goto Abort;
	if (mFile.Read(&atom.type, 4) != 4) goto Abort;

	bswap32(sz);

	if (sz == 0)
	{
		// Atom size is the remainder of the file.
		sz = mFile.Size() - atom.offset;

		// Size must be at least 8 bytes.
		if (sz < 8) goto Abort;
	}
	else if (sz == 1)
	{
		// Atom size is a 64-bit number.
		if (mFile.Read(&sz, 8) != 8) goto Abort;
		bswap64(sz);

		// Size must be at least 16 bytes and not extend past EOF.
		if (sz > (mFile.Size() - atom.offset))
		{
			sz = mFile.Size() - atom.offset;
		}
		if (sz < 16) goto Abort;
	}
	else
	{
		// Size must be at least 8 bytes and not extend past EOF.
		if (sz > (mFile.Size() - atom.offset))
		{
			sz = mFile.Size() - atom.offset;
		}
		if (sz < 8) goto Abort;
	}

	atom.size = sz;
	ret = true;

Abort:
	return ret;
}


bool InputFileQT::NextMoovAtom(QTAtom& atom)
{
	// Return the next atom in the 'moov' resource

	bool ret = false;
	__int64 sz = 0;

	atom.offset = moov_pos;
	if (moov_read(&sz, 4) != 4) goto Abort;
	if (moov_read(&atom.type, 4) != 4) goto Abort;

	bswap32(sz);

	if (sz == 0)
	{
		// Atom size is the remainder of the resource.
		sz = moov_size - atom.offset;

		// Size must be at least 8 bytes.
		if (sz < 8) goto Abort;
	}
	else if (sz == 1)
	{
		// Atom size is a 64-bit number.
		if (moov_read(&sz, 8) != 8) goto Abort;
		bswap64(sz);

		// Size must be at least 16 bytes and not extend past EOF.
		if (sz > (moov_size - atom.offset))
		{
			sz = moov_size - atom.offset;
		}
		if (sz < 16) goto Abort;
	}
	else
	{
		// Size must be at least 8 bytes and not extend past EOF.
		if (sz > (moov_size - atom.offset))
		{
			sz = moov_size - atom.offset;
		}
		if (sz < 8) goto Abort;
	}

	atom.size = sz;
	ret = true;

Abort:
	return ret;
}


int InputFileQT::ReadWholeSample(void *lpBuffer, const QTSampleInfo& si)
{
	// Read one whole sample from the file.
	// Caller is responsible for ensuring his buffer is large enough!

	if (lpBuffer != NULL && mFile.Seek(si.pos))
	{
		return mFile.Read(lpBuffer, si.size & (~QT_KEYFRAME));
	}
	return 0;
}


int InputFileQT::ReadRawStream(void *lpBuffer, __int64 bytepos, int bytes, QTTrack& qtt)
{
	// Read from an absolute byte position in the audio/video stream.

	// Uses the stream's tracking variables dwCurBlock and i64bytepos
	// to remember the last read position. Except for (re)initializing
	// to zero, no other code should ever modify these variables!

	if (lpBuffer != NULL && bytes > 0
		&& qtt.dwCurBlock < qtt.num_samples)
	{
		int offset;
		unsigned char *dst = (unsigned char *)lpBuffer;

		if (bytepos < qtt.i64bytepos)
		{
			// Requested data is before the current position
			for (;;)
			{
				if (qtt.dwCurBlock == 0)
				{
					// We are at the first block (byte position 0).
					if (bytepos < qtt.i64bytepos) return 0;
					break;
				}

				// Check the previous block position.
				--qtt.dwCurBlock;
				qtt.i64bytepos -= (qtt.sample_list[qtt.dwCurBlock].size & (~QT_KEYFRAME));
				if (bytepos >= qtt.i64bytepos) break;
			}
		}
		else
		{
			// Requested data is after the current position
			for (;;)
			{
				// Next position is our current position plus current block size.
				__int64 iNextPos = qtt.i64bytepos + (qtt.sample_list[qtt.dwCurBlock].size & (~QT_KEYFRAME));

				if (qtt.dwCurBlock == (qtt.num_samples - 1))
				{
					// We are at the last block (end of stream).
					if (bytepos >= iNextPos) return 0;
					break;
				}

				// Check the next block position.
				if (bytepos < iNextPos) break;
				qtt.i64bytepos = iNextPos;
				++qtt.dwCurBlock;
			}
		}

		// Starting offset in the current block
		offset = (int)(bytepos - qtt.i64bytepos);

		while (bytes > 0)
		{
			// Read the whole block if there is room.
			int blksize = qtt.sample_list[qtt.dwCurBlock].size & (~QT_KEYFRAME);
			int remains = blksize - offset;
			if (remains > bytes) remains = bytes;

			if (!mFile.Seek(qtt.sample_list[qtt.dwCurBlock].pos + offset)) break;
			if (mFile.Read(dst, remains) != remains) break;

			// Update destination buffer.
			dst   += remains;
			bytes -= remains;
			if (bytes == 0) break;	// finished

			// If this was the last block, stop.
			if (qtt.dwCurBlock >= (qtt.num_samples - 1)) break;

			// Go to the next block and start at offset 0.
			qtt.i64bytepos += blksize;
			++qtt.dwCurBlock;
			offset = 0;
		}

		// Return the actual number of bytes read.		
		return (int)(dst - (unsigned char *)lpBuffer);
	}

	return 0;
}


unsigned long InputFileQT::LowercaseFormat(unsigned long tag) const
{
	// Return a lowercase format tag
	char ch[4];
	*(unsigned long *)ch = tag;

	for (int i = 0; i < sizeof(ch); ++i)
	{
		if (ch[i] >= 'A' && ch[i] <= 'Z') ch[i] += ('a' - 'A');
	}

	return *(unsigned long *)ch;
}


unsigned long InputFileQT::UppercaseFormat(unsigned long tag) const
{
	// Return an uppercase format tag
	char ch[4];
	*(unsigned long *)ch = tag;

	for (int i = 0; i < sizeof(ch); ++i)
	{
		if (ch[i] >= 'a' && ch[i] <= 'z') ch[i] -= ('a' - 'A');
	}

	return *(unsigned long *)ch;
}


int InputFileQT::TransformAVCConfRec(const unsigned char *src, int srclen, unsigned char *dst, int &len) const
{
	// AVCDecoderConfigurationRecord,
	// ISO/IEC 14496-15, section 5.2.4.1.1

	// Transform the AVCDecoderConfigurationRecord into something
	// that ffdshow can understand by unwrapping it into raw NALUs.
	// Return the number of bytes in the result.  We have to save
	// this data and prepend it to all key frames of the AVC stream.
	// The "len" variable is used to return the lengthSize value.

	int n;
	int twice;
	int total = 0;

	if (srclen < 6) goto Abort;

	// 1 byte, configurationVersion (1)
	n = *src++;
	if (n != 1) goto Abort;

	// 1 byte, AVCProfileIndication
	// 1 byte, profile_compatibility
	// 1 byte, AVCLevelIndication
	src += 3;

	// 111111..  reserved
	// ......xx  lengthSizeMinusOne
	len = (*src++ & 3) + 1;

	// 111.....  reserved
	// ...xxxxx  numOfSequenceParameterSets
	n = *src++ & 0x1F;

	srclen -= 6;

	for (twice = 0; twice < 2; ++twice)
	{
		while (n-- > 0 && srclen > 0)
		{
			// sequenceParameterSetLength,
			// or pictureParameterSetLength
			int size = *src++;
			size = (size << 8) + *src++;
			
			if (size > 0)
			{
				if (size > srclen) size = srclen;

				// sequenceParameterSetNALUnit,
				// or pictureParameterSetNALUnit
				if (dst != NULL)
				{
					*(unsigned long *)dst = 0x01000000;
					memcpy(dst + 4, src, size);
					dst += 4 + size;
				}
				total  += 4 + size;
				src    += size;
				srclen -= size;
			}
		}

		if (srclen <= 0) break;

		// numOfPictureParameterSets
		n = *src++;
	}

Abort:
	return total;
}


/*
EsdsBox (stream specific description box),
usually holds the Bitrate/No of Channels 

It contains a number of (possibly optional?) sections (section 3 - 6)
(containing optional filler) with different info in each section. 

-> 4 bytes version/flags = 8-bit hex version + 24-bit hex flags (current = 0)

Section 3
-> 1 byte ES descriptor type tag = 8-bit hex value 0x03
-> 3 bytes optional extended descriptor type tag string = 3 * 8-bit hex value
   - types are 0x80,0x81,0xFE
-> 1 byte descriptor type length = 8-bit unsigned length
-> 2 bytes ES ID = 16-bit unsigned value
-> 1 byte stream priority = 8-bit unsigned value
   - Defaults to 16 and ranges from 0 through to 31

Section 4
-> 1 byte decoder config descriptor type tag = 8-bit hex value 0x04
-> 3 bytes optional extended descriptor type tag string = 3 * 8-bit hex value
   - types are 0x80,0x81,0xFE
-> 1 byte descriptor type length = 8-bit unsigned length
-> 1 byte object type ID = 8-bit unsigned value
-> 6 bits stream type = 3/4 byte hex value
   - type IDs are object descript. = 1 ; clock ref. = 2
   - type IDs are scene descript. = 4 ; visual = 4
   - type IDs are audio = 5 ; MPEG-7 = 6 ; IPMP = 7
   - type IDs are OCI = 8 ; MPEG Java = 9
   - type IDs are user private = 32
-> 1 bit upstream flag = 1/8 byte hex value
-> 1 bit reserved flag = 1/8 byte hex value set to 1
-> 3 bytes buffer size = 24-bit unsigned value
-> 4 bytes maximum bit rate = 32-bit unsigned value
-> 4 bytes average bit rate = 32-bit unsigned value 

Section 5
-> 1 byte decoder specific descriptor type tag 8-bit hex value 0x05
-> 3 bytes optional extended descriptor type tag string = 3 * 8-bit hex value
   - types are 0x80,0x81,0xFE
-> 1 byte descriptor type length = 8-bit unsigned length
-> 1 byte Audio profile Id - 5 bits Profile Id - 3 bits Unknown
-> 8 bits other flags - 3 bits unknown - 2 bits is No of Channels - 3 bits unknown

Section 6 
-> 1 byte SL config descriptor type tag = 8-bit hex value 0x06
-> 3 bytes optional extended descriptor type tag string = 3 * 8-bit hex value
   - types are 0x80,0x81,0xFE
-> 1 byte descriptor type length = 8-bit unsigned length
-> 1 byte SL value = 8-bit hex value set to 0x02 
*/

void InputFileQT::parse_esds_name(int max)
{
	// Skip the optional descriptor name if present.
	// Caller specifies the maximum number of chars to skip.

	// These are the only characters allowed in a name:
	const unsigned char valid[3] = { 0x80, 0x81, 0xFE };

	if (moov_pos < moov_size)
	{
		int i, k;

		for (i = 0; i < max; ++i)
		{
			unsigned char n = moov_bytes[moov_pos];

			for (k = 0; k < sizeof(valid); ++k)
			{
				if (n == valid[k]) break;
			}
			if (k == sizeof(valid)) break;

			if (++moov_pos == moov_size) break;
		}
	}
}


/*
ObjectTypeIndication, Description
0x00 Reserved for ISO use
0x01 Systems 14496-1 Simple Scene Description
0x02 Systems 14496-1 2D Scene Description
0x03 Interaction Stream MPEG-4 
0x04 Extended BIFS (h) MPEG-4 
0x05 AFX Stream (i) MPEG-4 
0x06 Font Data Stream MPEG-4 
0x07 Synthetised Texture MPEG-4 
0x08 Text Stream MPEG-4 
0x09 LASeR Stream MPEG-4 
0x0A Simple Aggregation Format (SAF) Stream MPEG-4 
0x20 Visual ISO/IEC 14496-2 (c) MPEG-4 
0x21 Visual ITU-T Recommendation H.264 | ISO/IEC 14496-10 (g) MPEG-4 
0x22 Parameter Sets for ITU-T Recommendation H.264 | ISO/IEC 14496-10 (g) MPEG-4 
0x40 Audio ISO/IEC 14496-3 (d) MPEG-4 
0x60 Visual ISO/IEC 13818-2 Simple Profile MPEG-4 
0x61 Visual ISO/IEC 13818-2 Main Profile MPEG-4 
0x62 Visual ISO/IEC 13818-2 SNR Profile MPEG-4 
0x63 Visual ISO/IEC 13818-2 Spatial Profile MPEG-4 
0x64 Visual ISO/IEC 13818-2 High Profile MPEG-4 
0x65 Visual ISO/IEC 13818-2 422 Profile MPEG-4 
0x66 Audio ISO/IEC 13818-7 Main Profile MPEG-4 
0x67 Audio ISO/IEC 13818-7 LowComplexity Profile MPEG-4 
0x68 Audio ISO/IEC 13818-7 Scaleable Sampling Rate Profile MPEG-4 
0x69 Audio ISO/IEC 13818-3 MPEG-4 
0x6A Visual ISO/IEC 11172-2 MPEG-4 
0x6B Audio ISO/IEC 11172-3 MPEG-4 
0x6C Visual ISO/IEC 10918-1 MPEG-4 
0x6D Portable Network Graphics (f) PNG 
0x6E Visual ISO/IEC 15444-1 (JPEG 2000) MPEG-4 
0xA0 EVRC Voice 3GPP2 
0xA1 SMV Voice 3GPP2 
0xA2 3GPP2 Compact Multimedia Format (CMF) 3GPP2 
0xA3 SMPTE VC-1 Video SMPTE 
0xA4 Dirac Video Coder Dirac 
0xA5 AC-3 Audio ETSI 
0xA6 Enhanced AC-3 audio ETSI 
0xA7 DRA Audio DRA 
0xA8 ITU G.719 Audio ITU G.719 
0xA9 DTS Coherent Acoustics audio DTS 
0xAA DTS-HD High Resolution Audio DTS 
0xAB DTS-HD Master Audio DTS 
0xAC DTS Express low bit rate audio, also known as DTS LBR DTSE 
0xC0 - 0xE0 user private  
0xE1 13K Voice 3GPP2 
0xE2 - 0xFE user private  
0xFF no object type specified (e) MPEG-4 

0x1F Vorbis?
0x33 Vorbis?
0x44 Vorbis?
*/

void InputFileQT::parse_esds(SDExtensions &ex, __int64 limit)
{
	// Elementary Stream Descriptor.
	// cf. ISO/IEC 14496-1, ISO/IEC 14496-14

	// Starts with ES_Descriptor, which typically
	// contains the other descriptors within it.
	// The codec data we want is in Section 5.

	int i;

	ex.ObjectTypeID = 0;
	moov_read(&i, 4);	// version (1 byte), flags (3 bytes)

	if (i == 0)
	{
		while (moov_pos < limit)
		{
			i = 0;
			moov_read(&i, 1);		// descriptor type tag

			if (i == 3)				// ES_Descriptor
			{
				// ES_Descriptor
				// 0 to 3 bytes, name (optional)
				// 1 byte, size
				// 2 bytes, ES_ID
				// 1 byte, flags / priority:
				// x.......  Stream dependence flag
				// .x......  URL flag
				// ..x.....  OCR stream flag
				// ...xxxxx  Stream priority

				parse_esds_name(3); // descriptor name (optional)
				moov_pos += 1;		// descriptor size
				moov_pos += 2;		// descriptor ES ID
				moov_pos += 1;		// stream flags / priority
			}
			else if (i == 4)
			{
				// DecoderConfigDescriptor
				// 0 to 3 bytes, name (optional)
				// 1 byte, size
				// 1 byte, object type indication
				// 1 byte, stream type, flags:
				// xxxxxx..  streamType
				// ......x.  Upstream flag
				// .......1  Reserved
				// 3 bytes, bufferSizeDB
				// 4 bytes, maxBitrate
				// 4 bytes, avgBitrate

				parse_esds_name(3); // descriptor name (optional)
				moov_pos += 1;		// descriptor size
				moov_read(&ex.ObjectTypeID, 1);
				moov_pos += 1;		// stream type / flags
				moov_pos += 3;		// buffer size
				moov_pos += 4;		// maximum bit rate
				moov_pos += 4;		// average bit rate
			}
			else if (i == 5)
			{
				// DecoderSpecificInfo
				// 0 to 3 bytes, name (optional)
				// 1 byte, size
				// n bytes, codec data

				if (ex.pCodecPrivate == NULL)
				{
					parse_esds_name(3); // descriptor name (optional)
					ex.iCodecPrivateSize = 0;
					moov_read(&ex.iCodecPrivateSize, 1);	// descriptor size
					if (ex.iCodecPrivateSize > 0)
					{
						ex.pCodecPrivate = moov_bytes + moov_pos;
					}
				}
				break;
			}
		}
	}
}


void InputFileQT::parse_wave(SDExtensions &ex, __int64 limit)
{
	// Parse a 'wave' container atom.
	// It may contain any of the following:
	// 'esds', 'frma', 'mp4a', 'enda'...

	QTAtom atom;

	while (moov_pos < limit)
	{
		if (!NextMoovAtom(atom)) break;
		if ((atom.offset + atom.size) > limit) break;

		if (atom.type == 'sdse')
		{
			parse_esds(ex, atom.offset + atom.size);
		}
		else if (atom.type == 'adne' && ex.iEndianness == 0)
		{
			int i = 0;
			moov_read(&i, 2);
			bswap16(i);
			ex.iEndianness = i;
		}
		else if (atom.type == A_MS11 && ex.pCodecPrivate == NULL)
		{
			// Microsoft IMA ADPCM WAVEFORMATEX
			moov_pos += FIELD_OFFSET(WAVEFORMATEX, cbSize);
			ex.iCodecPrivateSize = 0;
			moov_read(&ex.iCodecPrivateSize, 2);
			if (ex.iCodecPrivateSize > 0)
			{
				ex.pCodecPrivate = moov_bytes + moov_pos;
			}
		}

		moov_pos = (unsigned long)(atom.offset + atom.size);
	}
}

bool InputFileQT::parse_exts(SDExtensions& ex, __int64 limit)
{
	// Parse all known extensions to the 'stsd' atom.
	// A return value of false means insufficient memory.

	QTAtom atom;

	while (moov_pos < limit)
	{
		if (!NextMoovAtom(atom)) break;
		if ((atom.offset + atom.size) > limit) break;

		if (atom.type == '3cad')
		{
			// AC-3 Descriptor
			// The.Talk.Jackie.Evancho.June.24.2011.x264.720p60.mp4
			// has: 0x10 0x3D 0xC0 0x00 0x00 0x00 0x18

			static const unsigned char chan[8] = { 2, 1, 2, 3, 3, 4, 4, 5 };
			unsigned short buf;

			moov_read(&buf, 2);

			//   LSB      MSB
			// xx...... ........  fscod
			// ..xxxxx. ........  bsid
			// .......x xx......  bsmod
			// ........ ..xxx...  acmod
			// ........ .....x..  lfeon
			// ........ ......xx  frmsizcod[4..3]

			ex.wAC3Channels = chan[(buf >> 11) & 7] + ((buf >> 10) & 1);
		}
		else if (atom.type == 'psap')
		{
			// Pixel Aspect Ratio
			unsigned long h, v;

			moov_read(&h, 4);
			moov_read(&v, 4);

			if (h != 0 && v != 0)
			{
				bswap32(h);
				ex.hSpacing = h;
				bswap32(v);
				ex.vSpacing = v;
			}
		}
		else if (atom.type == 'Ccva' && ex.avc_dcr == NULL)
		{
			// AVC DecoderConfigurationRecord
			// cf. ISO/IEC 14496-15
			int len;

			int i = TransformAVCConfRec(
				moov_bytes + moov_pos,
				moov_size - moov_pos,
				NULL, len);

			if (i > 0)
			{
				// "i" is the total size needed for the DCR
				// "len" is the lengthSize value

				AVCSequence seq;

				ex.avc_dcr  = new unsigned char[i];
				if (ex.avc_dcr == NULL) return false;

				ex.dcr_size = i;
				ex.dcr_len  = len;

				TransformAVCConfRec(
					moov_bytes + moov_pos,
					moov_size - moov_pos,
					ex.avc_dcr, len);

				// At this point get the true AVC width/height
				if (seq.ParseSPS(ex.avc_dcr + 4, ex.dcr_len - 4))
				{
					ex.DisplayWidth  = seq.Width();
					ex.DisplayHeight = seq.Height();
				}
			}
		}
		else if (atom.type == 'evaw')
		{
			// 'wave' is itself a container, aargh...
			parse_wave(ex, atom.offset + atom.size);
		}
		else if (atom.type == 'sdse')
		{
			parse_esds(ex, atom.offset + atom.size);
		}
		else if (atom.type == 'lblg' && ex.pCodecPrivate == NULL)
		{
			// This entire atom is codec data for Microsoft VC-1
			ex.iCodecPrivateSize = (int)((atom.offset + atom.size) - moov_pos);
			if (ex.iCodecPrivateSize > 0)
				ex.pCodecPrivate = moov_bytes + moov_pos;
			else
				ex.iCodecPrivateSize = 0;
		}
		else if (atom.type == '1cvd' && ex.pCodecPrivate == NULL)
		{
			// Yet another VC-1 specific atom, mostly unknown.
			// 7 bytes (example: C6 60 24 00 12 4F 80), then data
			int i = 0;
			moov_read(&i, 1);	// profile?
			if ((i & 0xF0) == 0xC0)
			{
				moov_pos += 6;	// unknown
				ex.iCodecPrivateSize = (int)((atom.offset + atom.size) - moov_pos);
				if (ex.iCodecPrivateSize > 0)
					ex.pCodecPrivate = moov_bytes + moov_pos;
				else
					ex.iCodecPrivateSize = 0;
			}
		}

		moov_pos = (unsigned long)(atom.offset + atom.size);
	}

	return true;
}


unsigned long InputFileQT::RemapDefault(unsigned long dwFormat, unsigned long dwObjectID) const
{
	// Remap the QuickTime FOURCC based on known and tested samples.

	unsigned long dwRemap;

	switch (dwFormat)
	{
		// chad_vader.mp4
		case V_AVC1: dwRemap = '1CVA'; break;

		// psa.mov
		case V_CVID: dwRemap = 'divc'; break;

		// qt_fast_audio(dvc ).mov
		case V_DVC_: dwRemap = 'CVDC'; break;

		// qt_dv_pal_test.mov
		case V_DVCP: dwRemap = 'CVDP'; break;

		// h263.mov
		case V_H263: dwRemap = '362H'; break;

		// mario001.mov (doesn't work)
		case V_IV41: dwRemap = '14VI'; break;

		// Olympus 8080C.mov
		case V_JPEG: dwRemap = 'GEPJ'; break;

		// vb1(mjpa,raw ).mov
		case V_MJPA: dwRemap = 'APJM'; break;

		// mjpegb.mov (doesn't work)
		case V_MJPB: dwRemap = 'BPJM'; break;

		// sgi(mjpg).mov
		case V_MJPG: dwRemap = 'GPJM'; break;

		case V_MP4V:
		switch (dwObjectID)
		{
			// 100_0423.mov
			case 0x20:  // Visual ISO/IEC 14496-2 MPEG-4
				dwRemap = 'V4PM';   // 'DIVX';
				break;

			// 273G0147_01.MP4
			case 0x60: case 0x61: case 0x62:
			case 0x63: case 0x64: case 0x65:
				dwRemap = '2GPM';
				break;

			// Original PAR.mp4
			case 0x6A:  // Visual ISO/IEC 11172-2 MPEG-1
				dwRemap = '1GPM';
				break;

			// vc-1.mp4
			case 0xA3:  // SMPTE/Microsoft VC-1 Video
				dwRemap = '1CVW';
				break;
		}
		break;

		// ntsc_dv_smpte_bars.mov
		case V_RLE_: dwRemap = 'ELRQ'; break;

		// clip1(rpza, raw).mov
		case V_RPZA: dwRemap = 'AZPR'; break;

		// Strezz.3gp
		case V_S263: dwRemap = '362S'; break;

		// bs.mov
		case V_SVQ1: dwRemap = '1QVS'; break;

		// dichotomy.mov
		case V_SVQ3: dwRemap = '3QVS'; break;

		// vc-1.mov
		case V_VC_1: dwRemap = '1CVW'; break;

		default:
			dwRemap = UppercaseFormat(dwFormat);
			break;
	}

	return dwRemap;
}


bool InputFileQT::parse_vids(QTTrack& qtt, int size)
{
	// Parse QuickTime Video Sample Description
	ImageDescription    id;
	SDExtensions        ex;
	BITMAPINFOHEADER*   bmih;
	unsigned char*      origin;

	// Starts with ImageDescription header
	if (size < sizeof(id)) goto Abort;
	moov_read(&id, sizeof(id));
	size -= sizeof(id);

	qtt.dwFormatQT = id.dwDataFormat;
	qtt.dwFormat   = LowercaseFormat(id.dwDataFormat);

_RPT4(_CRT_WARN, "Video format: %c%c%c%c\n", qtt.dwFormat & 0xFF, (qtt.dwFormat >> 8) & 0xFF, (qtt.dwFormat >> 16) & 0xFF, qtt.dwFormat >> 24);

	// TODO: "MPEGSolution_jurassic.mp4" has empty ImageDescription!

	// Sample Description Extensions may follow
	// Save the current position before parsing.
	origin = moov_bytes + moov_pos;

	if (!parse_exts(ex, moov_pos + size))
	{
		// Fatal error
		mContext.mpCallbacks->SetErrorOutOfMemory();
		goto Abort;
	}

	// Copy PAR data if any.
	qtt.hSpacing = ex.hSpacing;
	qtt.vSpacing = ex.vSpacing;

	// Format specific video initialization
	switch (qtt.dwFormat)
	{
	case V_AVC1:	
		// Did we find the AVC Decoder Configuration Record?
		if (ex.avc_dcr == NULL)
		{
			log.Add("Video track %lu (AVC) is missing the Decoder Configuration Record.\r\n", qtt.dwTrackID);
			qtt.dwFlags |= QTT_UNSUPPORTED;
			goto Abort;
		}
		delete[] qtt.avc_dcr;
		qtt.avc_dcr  = ex.avc_dcr;
		ex.avc_dcr   = NULL;
		qtt.dcr_size = ex.dcr_size;
		qtt.avc_len  = ex.dcr_len;
		break;

	case V_SVQ3:
		// The Sorenson 3 decoder needs both ImageDescription
		// and 'SMI ' atom appended to the BITMAPINFOHEADER.
		ex.iCodecPrivateSize = sizeof(id) + size;
		ex.pCodecPrivate     = origin - sizeof(id);
		break;

	case V_MP4V:
#ifdef _DEBUG
		log.Add("Video track %lu ex.ObjectTypeID = %02X.\r\n", qtt.dwTrackID, ex.ObjectTypeID);
#endif
		switch (ex.ObjectTypeID)
		{
			case 0x20:  // Visual ISO/IEC 14496-2 MPEG-4

// TODO: "MPEG4 by philips.mp4" has incorrect width/height values
// in the header. Maybe we can parse the correct values from the VOL?

				// Did we find a VOL prefix?
				if (ex.pCodecPrivate == NULL || ex.iCodecPrivateSize <= 0)
				{
					log.Add("Video track %lu (MPEG-4 Part 2) is missing the VOL prefix.\r\n", qtt.dwTrackID);
					qtt.dwFlags |= QTT_UNSUPPORTED;
					goto Abort;
				}
				// If yes, store it in the avc_dcr (hack!)
				delete[] qtt.avc_dcr;
				qtt.avc_dcr = new unsigned char[ex.iCodecPrivateSize];
				if (qtt.avc_dcr == NULL)
				{
					mContext.mpCallbacks->SetErrorOutOfMemory();
					goto Abort;
				}
				memcpy(qtt.avc_dcr, ex.pCodecPrivate, ex.iCodecPrivateSize);
				qtt.dcr_size = ex.iCodecPrivateSize;

				ex.iCodecPrivateSize = 0;
				ex.pCodecPrivate = NULL;
				break;

			case 0x60:  // Visual ISO/IEC 13818-2 Simple Profile MPEG-2
			case 0x61:  // Visual ISO/IEC 13818-2 Main Profile MPEG-2
			case 0x62:  // Visual ISO/IEC 13818-2 SNR Profile MPEG-2
			case 0x63:  // Visual ISO/IEC 13818-2 Spatial Profile MPEG-2
			case 0x64:  // Visual ISO/IEC 13818-2 High Profile MPEG-2
			case 0x65:  // Visual ISO/IEC 13818-2 4:2:2 Profile MPEG-2
				break;

			case 0x6A:  // Visual ISO/IEC 11172-2 MPEG-1
				break;

			case 0xA3:  // SMPTE/Microsoft VC-1 Video
				if (ex.pCodecPrivate == NULL || ex.iCodecPrivateSize <= 0)
				{
					log.Add("Video track %lu (VC-1) is missing the codec private data.\r\n", qtt.dwTrackID);
					qtt.dwFlags |= QTT_UNSUPPORTED;
					goto Abort;
				}
				break;

			default:
				log.Add("Video track %lu has an unknown Elementary Stream Descriptor Object Type ID: 0x%02X.\r\n", qtt.dwTrackID, ex.ObjectTypeID);
				qtt.dwFlags |= QTT_UNSUPPORTED;
				goto Abort;
		}
	}

	// Now construct a BITMAPINFOHEADER including
	// the codec private data (if any).

	qtt.pFormat = new unsigned char[sizeof(BITMAPINFOHEADER) + ex.iCodecPrivateSize];
	if (qtt.pFormat == NULL)
	{
		mContext.mpCallbacks->SetErrorOutOfMemory();
		goto Abort;
	}
	qtt.lFormatLen = sizeof(BITMAPINFOHEADER) + ex.iCodecPrivateSize;
	memset(qtt.pFormat, 0, qtt.lFormatLen);

	bmih = (BITMAPINFOHEADER *)qtt.pFormat;
	bmih->biSize = qtt.lFormatLen;

	if (ex.DisplayWidth > 0 && ex.DisplayHeight > 0)
	{
		// Width and height derived from AVC sequence parameter set
		bmih->biWidth  = ex.DisplayWidth;
		bmih->biHeight = ex.DisplayHeight;
	}
	else
	{
		bmih->biWidth = id.wWidth;
		bswap16(bmih->biWidth);
		bmih->biHeight = id.wHeight;
		bswap16(bmih->biHeight);
	}

	bmih->biBitCount = id.wDepth;
	bswap16(bmih->biBitCount);
	bmih->biPlanes = 1;

	bmih->biSizeImage = ((bmih->biWidth + 3) & -4) * bmih->biHeight * 3;

	// Check if we need to remap the FOURCC
	bmih->biCompression = pOptions->RemapFourcc(qtt.dwFormat);
	if (bmih->biCompression == qtt.dwFormat)
	{
		// No remap override took place, so use a default remapping:
		bmih->biCompression = RemapDefault(qtt.dwFormat, ex.ObjectTypeID);

		// Check again for override of default remapping:
		if (bmih->biCompression != qtt.dwFormat)
		{
			bmih->biCompression = pOptions->RemapFourcc(bmih->biCompression);
		}
	}

	// Append the codec private data if any
	if (ex.pCodecPrivate != NULL && ex.iCodecPrivateSize > 0)
	{
		memcpy(qtt.pFormat + sizeof(BITMAPINFOHEADER), ex.pCodecPrivate, ex.iCodecPrivateSize);
	}

	return true;

Abort:
	return false;
}


bool InputFileQT::parse_auds(QTTrack& qtt, int size)
{
	// Parse QuickTime Audio Sample Description
	SoundDescription    sd;
	SDExtensions        ex;

	// Starts with SoundDescription header
	if (size < sizeof(sd)) goto Abort;
	moov_read(&sd, sizeof(sd));
	size -= sizeof(sd);

	qtt.dwFormatQT = sd.dwDataFormat;
	qtt.dwFormat   = LowercaseFormat(sd.dwDataFormat);

_RPT4(_CRT_WARN, "Audio format: %c%c%c%c\n", qtt.dwFormat & 0xFF, (qtt.dwFormat >> 8) & 0xFF, (qtt.dwFormat >> 16) & 0xFF, qtt.dwFormat >> 24);

	// Check the version
	bswap16(sd.wVersion);

	if (sd.wVersion == 1 && size >= 16)
	{
		// Version 1 has 4 extra fields comprising 16 bytes.
		// Note: for ima4 these are:
		// 64, 34, 34, 2
		moov_read(&qtt.dwSamplesPerPacket, 4);
		moov_read(&qtt.dwBytesPerPacket, 4);
		moov_read(&qtt.dwBytesPerFrame, 4);
		moov_read(&qtt.dwBytesPerSample, 4);
		size -= 16;

		bswap32(qtt.dwSamplesPerPacket);
		bswap32(qtt.dwBytesPerPacket);
		bswap32(qtt.dwBytesPerFrame);
		bswap32(qtt.dwBytesPerSample);
	}
	else if (sd.wVersion >= 2)
	{
		// Version 2 has 8 extra fields comprising 36 bytes.
		// Currently this version is not supported.
		log.Add("Audio track %lu has an unsupported Sound Description version (%hu).\r\n", qtt.dwTrackID, sd.wVersion);
		qtt.dwFlags |= QTT_UNSUPPORTED;
		goto Abort;
	}

	// Sample Description Extensions may follow
	if (!parse_exts(ex, moov_pos + size))
	{
		mContext.mpCallbacks->SetErrorOutOfMemory();
		goto Abort;
	}
	
	// Format specific audio initialization
	switch (qtt.dwFormat)
	{
	case A__MP3:
	case A_MS_U:
		ex.pCodecPrivate     = NULL;
		ex.iCodecPrivateSize = sizeof(MPEG1WAVEFORMAT) - sizeof(WAVEFORMATEX);
		if (ex.iCodecPrivateSize < MPEGLAYER3_WFX_EXTRA_BYTES)
			ex.iCodecPrivateSize = MPEGLAYER3_WFX_EXTRA_BYTES;
		// We fill in the extra data in PostInit().
		break;

	case A_MP4A:
#ifdef _DEBUG
		log.Add("Audio track %lu ex.ObjectTypeID = %02X.\r\n", qtt.dwTrackID, ex.ObjectTypeID);
#endif
		switch (ex.ObjectTypeID)
		{
			case 0x40:  // Audio ISO/IEC 14496-3 AAC
			case 0x66:  // Audio ISO/IEC 13818-7 Main Profile MPEG-4 
			case 0x67:  // Audio ISO/IEC 13818-7 LowComplexity Profile MPEG-4 
			case 0x68:  // Audio ISO/IEC 13818-7 Scaleable Sampling Rate Profile MPEG-4 
				if (ex.iCodecPrivateSize < 2)
				{
					log.Add("Audio track %lu (AAC) is missing the AudioSpecificConfig data.\r\n", qtt.dwTrackID);
					qtt.dwFlags |= QTT_UNSUPPORTED;
					goto Abort;
				}
				// Add ADTS headers to AAC?
				if (pOptions->AddADTSHeaders())
				{
					qtt.dwFlags |= QTT_ADD_ADTS;
				}
				break;

			case 0x69:  // Audio ISO/IEC 13818-3 MPEG-2
			case 0x6B:  // Audio ISO/IEC 11172-3 MPEG-1
				ex.pCodecPrivate     = NULL;
				ex.iCodecPrivateSize = sizeof(MPEG1WAVEFORMAT) - sizeof(WAVEFORMATEX);
				if (ex.iCodecPrivateSize < MPEGLAYER3_WFX_EXTRA_BYTES)
					ex.iCodecPrivateSize = MPEGLAYER3_WFX_EXTRA_BYTES;
				// We fill in the extra data in PostInit().
				break;

			default:
				log.Add("Audio track %lu has an unknown Elementary Stream Descriptor Object Type ID: 0x%02X.\r\n", qtt.dwTrackID, ex.ObjectTypeID);
				qtt.dwFlags |= QTT_UNSUPPORTED;
				goto Abort;
		}
		break;
		
	case A_MS11:
		if (ex.pCodecPrivate == NULL || ex.iCodecPrivateSize <= 0)
		{
			log.Add("Audio track %lu (Microsoft IMA ADPCM) is missing the wave format data.\r\n", qtt.dwTrackID);
			qtt.dwFlags |= QTT_UNSUPPORTED;
			goto Abort;
		}
		break;

	case A_FL32:
	case A_FL64:
	case A_IN24:
	case A_IN32:
		// Get the endianness
		qtt.iEndianness = ex.iEndianness;
		break;

	case A_AC_3:
		// Get the real channels from the 'dac3' atom
		if (ex.wAC3Channels != 0)
		{
			sd.wNumberOfChannels = ex.wAC3Channels;
			bswap16(sd.wNumberOfChannels);
		}
		break;

	case A_SAMR:
	case A_SAWB:
		// Add AMR header?
		if (pOptions->AddAMRHeader())
		{
			qtt.dwFlags |= QTT_ADD_AMR;
		}
		break;
	}

	// Now construct a WAVEFORMATEX including
	// the codec private data (if any).

	qtt.pFormat = new unsigned char[sizeof(WAVEFORMATEX) + ex.iCodecPrivateSize];
	if (qtt.pFormat == NULL)
	{
		mContext.mpCallbacks->SetErrorOutOfMemory();
		goto Abort;
	}
	qtt.lFormatLen = sizeof(WAVEFORMATEX) + ex.iCodecPrivateSize;
	memset(qtt.pFormat, 0, qtt.lFormatLen);

	if (qtt.dwFormat == A_MS11)
	{
		// Microsoft IMA ADPCM, copy the entire WAVEFORMATEX.
		memcpy(qtt.pFormat, ex.pCodecPrivate - sizeof(WAVEFORMATEX), qtt.lFormatLen);
	}
	else
	{
		// Construct the format and append private data if any.
		WAVEFORMATEX *wfex;
		wfex = (WAVEFORMATEX *)qtt.pFormat;

		wfex->nChannels = sd.wNumberOfChannels;
		bswap16(wfex->nChannels);

		wfex->wBitsPerSample = sd.wSampleSize;
		bswap16(wfex->wBitsPerSample);

		// This is 16:16 fixed point!
		wfex->nSamplesPerSec = sd.dwSampleRate;
		bswap32(wfex->nSamplesPerSec);
		if (wfex->nSamplesPerSec < 0xFFFF8000) wfex->nSamplesPerSec += 32768;
		wfex->nSamplesPerSec >>= 16;

		wfex->cbSize = ex.iCodecPrivateSize;
		if (ex.pCodecPrivate != NULL && ex.iCodecPrivateSize > 0)
		{
			memcpy(qtt.pFormat + sizeof(WAVEFORMATEX), ex.pCodecPrivate, ex.iCodecPrivateSize);
		}
	}

	return true;

Abort:
	return false;
}


void InputFileQT::parse_stbl(QTTrack& qtt, __int64 limit)
{
	// Parse all of the stream information.

	QTAtom atom;
	unsigned long i;

	while (moov_pos < limit)
	{
		if (!NextMoovAtom(atom)) break;
		if ((atom.offset + atom.size) > limit) break;

		if (atom.type == 'dsts' && qtt.pFormat == NULL)
		{
			// Sample Description

			moov_pos += 4;		// Version (1 byte), Flags (3 bytes)
			moov_read(&i, 4);	// Number of Entries (at least 1)
			if (i != 0)
			{
				int size;

				// Size of Sample Description
				moov_read(&size, 4);
				bswap32(size);

				if (qtt.fccType == streamtypeVIDEO)
				{
					if (!parse_vids(qtt, size)) goto Abort;
				}
				else if (qtt.fccType == streamtypeAUDIO)
				{
					if (!parse_auds(qtt, size)) goto Abort;
				}
			}
		}
		else if (atom.type == 'csts' && qtt.chunk_info == NULL)
		{
			// Sample-To-Chunk table

			moov_pos += 4;		// Version (1 byte), Flags (3 bytes)
			moov_read(&i, 4);	// Number of Entries
			if (i != 0)
			{
				bswap32(i);
				qtt.num_chunk_infos = i;
				qtt.chunk_info = (QTChunk *)(moov_bytes + moov_pos);

				for (i = 0; i < qtt.num_chunk_infos; ++i)
				{
					// Do not write out-of-bounds!
					moov_pos += sizeof(QTChunk);
					if (moov_pos > moov_size) break;
					bswap32(qtt.chunk_info[i].first);
					bswap32(qtt.chunk_info[i].samples);
					bswap32(qtt.chunk_info[i].ID);
				}
			}
		}
		else if (atom.type == 'ssts' && qtt.keyframe == NULL)
		{
			// Sync Samples (keyframes)

			moov_pos += 4;		// Version (1 byte), Flags (3 bytes)
			moov_read(&i, 4);	// Number of entries
			if (i != 0)
			{
				bswap32(i);
				qtt.num_keyframes = i;
				qtt.keyframe = (unsigned long *)(moov_bytes + moov_pos);

				for (i = 0; i < qtt.num_keyframes; ++i)
				{
					// Do not write out-of-bounds!
					moov_pos += sizeof(unsigned long);
					if (moov_pos > moov_size) break;
					bswap32(qtt.keyframe[i]);
				}
			}
		}
		else if (atom.type == 'octs' && qtt.chunk_offset32 == NULL && qtt.chunk_offset64 == NULL)
		{
			// Chunk Offsets (32-bit)

			moov_pos += 4;		// Version (1 byte), Flags (3 bytes)
			moov_read(&i, 4);	// Number of entries
			if (i != 0)
			{
				bswap32(i);
				qtt.num_chunk_offsets = i;
				qtt.chunk_offset32 = (unsigned long *)(moov_bytes + moov_pos);

				for (i = 0; i < qtt.num_chunk_offsets; ++i)
				{
					// Do not write out-of-bounds!
					moov_pos += sizeof(unsigned long);
					if (moov_pos > moov_size) break;
					bswap32(qtt.chunk_offset32[i]);
				}
			}
		}
		else if (atom.type == '46oc' && qtt.chunk_offset64 == NULL)
		{
			// Chunk Offsets (64-bit)

			moov_pos += 4;		// Version (1 byte), Flags (3 bytes)
			moov_read(&i, 4);	// Number of entries
			if (i != 0)
			{
				bswap32(i);
				qtt.num_chunk_offsets = i;
				qtt.chunk_offset32 = NULL;
				qtt.chunk_offset64 = (__int64 *)(moov_bytes + moov_pos);

				for (i = 0; i < qtt.num_chunk_offsets; ++i)
				{
					// Do not write out-of-bounds!
					moov_pos += sizeof(__int64);
					if (moov_pos > moov_size) break;
					bswap64(qtt.chunk_offset64[i]);
				}
			}
		}
		else if (atom.type == 'zsts' && qtt.sample_list == NULL)
		{
			// Sample Sizes

			moov_pos += 4;		// Version (1 byte), Flags (3 bytes)

			moov_read(&i, 4);	// Sample size
			bswap32(i);
			qtt.default_sample_size = i;

			moov_read(&i, 4);	// Number of entries
			bswap32(i);
			qtt.num_samples = i;

			if (qtt.default_sample_size == 0)
			{
				qtt.sample_size = (unsigned long *)(moov_bytes + moov_pos);

				for (i = 0; i < qtt.num_samples; ++i)
				{
					// Do not write out-of-bounds!
					moov_pos += sizeof(unsigned long);
					if (moov_pos > moov_size) break;
					bswap32(qtt.sample_size[i]);
				}
			}
		}
		else if (atom.type == 'stts' && qtt.time_to_sample_table == NULL)
		{
			// Time-To-Sample table

			moov_pos += 4;		// Version (1 byte), Flags (3 bytes)
			moov_read(&i, 4);	// Number of Entries
			if (i != 0)
			{
				bswap32(i);
				qtt.time_to_sample_count = i;
				qtt.time_to_sample_table = (QTTimeToSampleEntry *)(moov_bytes + moov_pos);

				for (i = 0; i < qtt.time_to_sample_count; ++i)
				{
					// Do not write out-of-bounds!
					moov_pos += sizeof(unsigned long) * 2;
					if (moov_pos > moov_size) break;
					bswap32(qtt.time_to_sample_table[i].count);
					bswap32(qtt.time_to_sample_table[i].duration);
				}
			}
		}

		moov_pos = (unsigned long)(atom.offset + atom.size);
	}

Abort:
	return;
}


void InputFileQT::parse_minf(QTTrack &qtt, __int64 limit)
{
	// Parse Media Information atom
	QTAtom atom;

	while (moov_pos < limit)
	{
		if (!NextMoovAtom(atom)) break;
		if ((atom.offset + atom.size) > limit) break;

		if (atom.type == 'lbts')
		{
			// Sample Tables
			parse_stbl(qtt, atom.offset + atom.size);
			break;
		}

		moov_pos = (unsigned long)(atom.offset + atom.size);
	}
}


void InputFileQT::parse_mdia(QTTrack& qtt, __int64 limit)
{
	// Parse Media atom
	QTAtom atom;
	int i;

	while (moov_pos < limit)
	{
		if (!NextMoovAtom(atom)) break;
		if ((atom.offset + atom.size) > limit) break;

		if (atom.type == 'dhdm')
		{
			// Media Header (required)
			unsigned long i = 0;

			moov_read(&i, 1);	// Version
			moov_pos += 3;		// Flags

			if (i == 1)
			{
				__int64 i64;

				moov_pos += 8;		// Creation time
				moov_pos += 8;		// Modification time

				moov_read(&i, 4);	// Time scale
				bswap32(i);
				qtt.dwTimeScale = i;

				moov_read(&i64, 8);	// Duration
				bswap64(i64);
				qtt.i64Duration = i64;
			}
			else
			{
				moov_pos += 4;		// Creation time
				moov_pos += 4;		// Modification time

				moov_read(&i, 4);	// Time scale
				bswap32(i);
				qtt.dwTimeScale = i;

				moov_read(&i, 4);	// Duration
				bswap32(i);
				qtt.i64Duration = i;
			}
			
//			moov_pos += 2;		// Language code
//			moov_pos += 2;		// Quality
		}
		else if (atom.type == 'rldh')
		{
			// Handler Reference

			moov_pos += 4;		// Version (1 byte), Flags (3 bytes)
			moov_read(&i, 4);	// Component type

			if (i == 'rlhm' || i == 0)
			{
				moov_read(&i, 4);	// Component subtype

				if (i == 'ediv')
				{
					qtt.fccType = streamtypeVIDEO;
				}
				else if (i == 'nuos')
				{
					qtt.fccType = streamtypeAUDIO;
				}

//				moov_pos += 4;		// Component manufacturer
//				moov_pos += 4;		// Component flags
//				moov_pos += 4;		// Component flags mask
//				moov_pos += 1;		// Component name (variable)
			}
		}
		else if (atom.type == 'fnim')
		{
			// Media Information
			parse_minf(qtt, atom.offset + atom.size);
		}

		moov_pos = (unsigned long)(atom.offset + atom.size);
	}
}


unsigned long InputFileQT::preparse_fmt(__int64 limit)
{
	// Hack; parse ahead to determine the Track data format
	QTAtom atom;

	unsigned long fmt = 0;
	unsigned long back = moov_pos;

	while (moov_pos < limit)
	{
		if (!NextMoovAtom(atom)) break;
		if ((atom.offset + atom.size) > limit) break;

		if (atom.type == 'aidm')
		{
			// Descend into Media atom
			limit = atom.offset + atom.size;
			continue;
		}

		if (atom.type == 'fnim')
		{
			// Descend into Media Information atom
			limit = atom.offset + atom.size;
			continue;
		}

		if (atom.type == 'lbts')
		{
			// Descend into Sample Tables atom
			limit = atom.offset + atom.size;
			continue;
		}

		if (atom.type == 'dsts')
		{
			// Sample Description, hooray!
			unsigned long i;

			moov_pos += 4;		// Version (1 byte), Flags (3 bytes)
			moov_read(&i, 4);	// Number of Entries
			if (i == 0) break;

			moov_read(&i, 4);	// size
			bswap32(i);
			if (i < 4) break;

			moov_read(&fmt, 4);	// data format
			fmt = LowercaseFormat(fmt);
			break;
		}

		moov_pos = (unsigned long)(atom.offset + atom.size);
	}

	moov_pos = back;	
	return fmt;
}


void InputFileQT::parse_trak(__int64 limit)
{
	// Parse one whole QuickTime Track.

	QTAtom atom;
	DWORD dwFormat;
	DWORD fccType;
	QTTrack *qtt = NULL;

	if (num_tracks >= MAX_TRACKS) goto Abort;

	// Get the format first so we know what kind of track it is.
	dwFormat = preparse_fmt(limit);

	// All audio tracks need special handlers.
	// Take the opportunity now to set the fccType, because
	// some files have the 'hdlr' atom after the 'stsd' atom!
	fccType = streamtypeAUDIO;

	switch (dwFormat)
	{
		case A_AC_3:
			qtt = new QTTrack_ac_3;
			break;
		case A_FL32:
		case A_FL64:
			qtt = new QTTrack_float;
			break;
		case A__MP3:
			qtt = new QTTrack__mp3;
			break;
		case A_IMA4:
			qtt = new QTTrack_ima4;
			break;
		case A_MP4A:
			qtt = new QTTrack_mp4a;
			break;
		case A_MS11:
			qtt = new QTTrack_ms11;
			break;
		case A_MS_U:
			qtt = new QTTrack_ms_u;
			break;
		case A_SAMR:
		case A_SAWB:
			qtt = new QTTrack_samr;
			break;
		case A_IN24:
		case A_IN32:
		case A_RAW_:
		case A_SOWT:
		case A_TWOS:
			qtt = new QTTrack_PCM;
			break;
		case A_ALAW:
		case A_ULAW:
			qtt = new QTTrack_ulaw;
			break;
		default:
			fccType = streamtypeVIDEO;
			qtt = new QTTrack;
			break;
	}

	if (qtt == NULL)
	{
		mContext.mpCallbacks->SetErrorOutOfMemory();
		goto Abort;
	}

	// Now fully parse all Track information
	qtt->fccType = fccType;

	while (moov_pos < limit)
	{
		if (!NextMoovAtom(atom)) break;
		if ((atom.offset + atom.size) > limit) break;

		if (atom.type == 'dhkt')
		{
			// Track Header (required)
			unsigned long i = 0;

			moov_read(&i, 1);	// Version
			moov_pos += 3;		// Flags

			if (i == 1)
			{
				moov_pos += 8;		// Creation time
				moov_pos += 8;		// Modification time
			}
			else
			{
				moov_pos += 4;		// Creation time
				moov_pos += 4;		// Modification time
			}

		// Track ID zero is supposed to be forbidden, but I have
		// seen them in the wild.  Enforce a nonzero Track ID!

			moov_read(&i, 4);	// Track ID
			bswap32(i);
			qtt->dwTrackID = i;

//			moov_pos += 4;		// Reserved
//			moov_pos += 4;		// Duration (8 bytes for version 1)
//			moov_pos += 8;		// Reserved
//			moov_pos += 2;		// Layer
//			moov_pos += 2;		// Alternate group
//			moov_pos += 2;		// Volume
//			moov_pos += 2;		// Reserved
//			moov_pos += 36;		// Matrix structure
//			moov_pos += 4;		// Track width
//			moov_pos += 4;		// Track height
		}
		else if (atom.type == 'aidm')
		{
			// Media atom (required)
			parse_mdia(*qtt, atom.offset + atom.size);
		}

		moov_pos = (unsigned long)(atom.offset + atom.size);
	}

	// Track is parsed, now verify it...
	if (QTTrack::Verify(*qtt))
	{
		// Hooray, we have a new Track!
		Track[num_tracks++] = qtt;
		return;
	}

Abort:
	// Track is not usable
	delete qtt;
	return;
}

/*
Problems:
Biggs.mov
cut1_tatooine.mov
MAC Hello Kitty.mp4
cakesong.mov
smaller size hope good.mov
pd_3_25_02.mov
Quinceanera.mov
*/

bool InputFileQT::parse_moov(const QTAtom& moov_atom)
{
	QTAtom atom;

	// Load the entire 'moov' resource into memory.

	moov_bytes = new unsigned char[(size_t)moov_atom.size];
	if (moov_bytes == NULL)
	{
		mContext.mpCallbacks->SetErrorOutOfMemory();
		goto Abort;
	}

	moov_size = (unsigned long)moov_atom.size;
	moov_pos  = 0;

	if (!mFile.Seek(moov_atom.offset)) goto Abort;
	if ((unsigned long)mFile.Read(moov_bytes, moov_size) != moov_size)
	{
		// Truncated 'moov' chunk!
		return false;
	}

	// Skip past the 'moov' header
	if (!NextMoovAtom(atom)) goto Abort;

	while (moov_pos < moov_size)
	{
		if (!NextMoovAtom(atom)) break;
		if ((atom.offset + atom.size) > moov_size) break;

		if (atom.type == 'vomc')
		{
			// Aargh, compressed movie atom!
			unsigned long ctype = 0;
			unsigned long csize = 0;
			unsigned char *moov_bytes2 = NULL;
			int retval;

			// Locate ctype ('zlib') and csize,
			// we have to find both!

			while (moov_pos < moov_size)
			{
				if (!NextMoovAtom(atom)) break;
				if ((atom.offset + atom.size) > moov_size) break;

				if (atom.type == 'mocd')
				{
					moov_read(&ctype, 4);
					if (csize) break;
				}
				else if (atom.type == 'dvmc')
				{
					moov_read(&csize, 4);
					bswap32(csize);
					if (ctype) break;
				}

				moov_pos = (unsigned long)(atom.offset + atom.size);
			}

			if (ctype != 'bilz') break;
			if (csize == 0) break;

			// Decompress, and reparse...
			moov_bytes2 = new unsigned char[csize];
			if (moov_bytes2 == NULL)
			{
				mContext.mpCallbacks->SetErrorOutOfMemory();
				break;
			}

			retval = uncompress(
				moov_bytes2, &csize,
				moov_bytes + moov_pos,
				moov_size  - moov_pos);

			if (retval != 0)
			{
				// Decompression failed!
				delete[] moov_bytes2;
				break;
			}
			
			delete[] moov_bytes;

			moov_bytes = moov_bytes2;
			moov_size  = csize;
			moov_pos   = 0;

			// Skip past the 'moov' header
			if (!NextMoovAtom(atom)) break;

			// Now reparse
			continue;
		}

		if (atom.type == 'kart')
		{
			parse_trak(atom.offset + atom.size);
		}

		moov_pos = (unsigned long)(atom.offset + atom.size);
	}

Abort:
	return (num_tracks > 0);
}


bool InputFileQT::PostInit()
{
	// Finish deriving info for all found tracks.

	int i;

	if (num_tracks < 1) goto Abort;

#if 0
	// For testing, remove any audio tracks
	i = 0;
	while (i < num_tracks) {
		if (Track[i]->fccType == streamtypeAUDIO) {
			int j;
			delete Track[i];
			for (j = i + 1; j < num_tracks; ++j) {
				Track[j - 1] = Track[j];
			}
			--num_tracks;
		} else {
			++i;
		}
	}
#endif

	// Create sample info array for all tracks.
	// Note: Tracks were already Verified in parse_trak!

	i = 0;
	while (i < num_tracks)
	{
		if (!Track[i]->Finish())
		{
			log.Add("Cannot finish %s track %lu, discarding.\r\n", (Track[i]->fccType == streamtypeVIDEO)? "video": "audio", Track[i]->dwTrackID);
			// Track is not usable, delete it.
			int k;
			delete Track[i];
			for (k = i + 1; k < num_tracks; ++k)
			{
				Track[k - 1] = Track[k];
			}
			Track[--num_tracks] = NULL;
			continue;
		}
		++i;
	}

	// Special case for MPEG Audio under the 'mp4a' tag.
	// We need to find out the Layer to create a proper
	// wave format.
	for (i = 0; i < num_tracks; ++i)
	{
		if (Track[i]->dwFormat == A_MP4A)
		{
			const WAVEFORMATEX *wfex = (WAVEFORMATEX *)Track[i]->pFormat;
			if (wfex->wFormatTag == WAVE_FORMAT_MPEGLAYER3)
			{
				unsigned int hdr;
				if (ReadRawStream(&hdr, 0, 4, *Track[i]) == 4)
				{
					QTTrack_mp4a *m = static_cast<QTTrack_mp4a*>(Track[i]);
					m->ParseHeader(hdr);
				}
			}
		}
	}

	// Log unsupported tracks if any.
	for (i = 0; i < num_tracks; ++i)
	{
		if (Track[i]->dwFlags & QTT_UNSUPPORTED)
		{
			log.Add("Unsupported %s track %i.\r\n", (Track[i]->fccType == streamtypeVIDEO)? "video": "audio", Track[i]->dwTrackID);
		}
	}

	if (num_tracks >= 1) return true;

Abort:
	return false;
}


const wchar_t *InputFileQT::FileTitle(const wchar_t *szFile) const
{
	// Return a pointer to the filename component.
	int i, last = 0;

	for (i = 0; ; ++i)
	{
		wchar_t wc = szFile[i];
		if (wc == L'\0') break;
		if (wc == L'\\' || wc == L'/') last = i + 1;
	}

	return (szFile + last);
}


bool InputFileQT::IsSupported_ftyp(unsigned long ftyp) const
{
	switch (ftyp)
	{
		case '  tq':
		case ' A4M':
		case ' V4M':
		case '4pmm':
		case 'b2g3':
		case '4pg3':
		case '5pg3':
		case '6pg3':
		case '24pm':
		case '1cva':
		case 'a2g3':
		case 'mosi':
		case 'ECAF':
		case ' v4f':
		case 'VNSM':
			return true;
	}
	return false;
}


void InputFileQT::Init(const wchar_t *szFile, IVDXInputOptions *opts)
{
	unsigned long i;
	QTAtom atom;
	bool success;

	num_tracks = 0;
	success = false;

#ifdef _DEBUG
	log.Add("Debugging...\r\n");
#endif

	if (!mFile.OpenW(szFile))
	{
		mContext.mpCallbacks->SetError("Cannot open file: %ls", FileTitle(szFile));
		return;
	}

	if (opts != NULL)
	{
		// Get options from VirtualDub
		pOptions = (InputOptionsQT *)opts;
	}
	else
	{
		// Get options (if any) from the registry
		pOptions = new InputOptionsQT;
		if (pOptions == NULL)
		{
			mContext.mpCallbacks->SetErrorOutOfMemory();
			return;
		}
	}
	pOptions->AddRef();

	while (NextAtom(atom))
	{
		if (atom.type == 'pytf')
		{
			unsigned char ftyp[5];
			bool supported;

			if (mFile.Read(ftyp, 4) != 4) goto Abort;
			if (!mFile.Move(4)) goto Abort;		// minor version

			major_brand = *(unsigned long *)ftyp;
			supported = IsSupported_ftyp(major_brand);

			if (!supported)
			{
				// Major brand is unknown; check compatible brands
				while (mFile.Pos() <= (atom.offset + atom.size - 4))
				{
					if (mFile.Read(&compat_brand, 4) != 4) goto Abort;
					supported = IsSupported_ftyp(compat_brand);
					if (supported) break;
				}
				
				if (!supported)
				{
					// No compatible brands are known.
					for (i = 0; i < 4; ++i)
					{
						if (ftyp[i] < 32) ftyp[i] = 32;
					}
					ftyp[4] = 0;

					mContext.mpCallbacks->SetError("Unsupported ftyp major_brand: '%s'", (char *)ftyp);
					return;
				}
			}
		}
		else if (atom.type == 'voom')
		{
			parse_moov(atom);
			break;
		}

		mFile.Seek(atom.offset + atom.size);
	}

	// Voila; the movie is parsed!

	// Now construct sample lookup tables
	// based on the QuickTime information

	success = PostInit();

Abort:
	if (!success)
	{
		mContext.mpCallbacks->SetError("Cannot parse %ls", FileTitle(szFile));
	}
}


bool InputFileQT::Append(const wchar_t *szFile)
{
	return false;
}


bool InputFileQT::PromptForOptions(VDXHWND hwndParent, IVDXInputOptions **ppOptions)
{
	InputOptionsQT *p = new InputOptionsQT;

	*ppOptions = p;
	if (p == NULL) return false;

	p->AddRef();
	p->OptionsDlg((HWND)hwndParent);

	return true;
}


bool InputFileQT::CreateOptions(const void *buf, uint32 len, IVDXInputOptions **ppOptions)
{
	InputOptionsQT *p = new InputOptionsQT;

	*ppOptions = p;
	if (p == NULL) return false;

	p->AddRef();
	p->Load(buf, len);

	return true;
}


void InputFileQT::DisplayInfo(VDXHWND hwndParent)
{
	DialogBoxParamA(g_hModule, MAKEINTRESOURCEA(IDD_QT_INFO),
	    (HWND)hwndParent, InfoDlgProc, (LPARAM)this);
}


bool InputFileQT::GetVideoSource(int index, IVDXVideoSource **ppVS)
{
	int i;

	*ppVS = NULL;

	for (i = 0; i < num_tracks; ++i)
	{
		if (Track[i] != NULL
			&& Track[i]->fccType == streamtypeVIDEO
			&& (Track[i]->dwFlags & QTT_UNSUPPORTED) == 0)
		{
			if (index == 0) break;
			--index;
		}
	}

	if (i < num_tracks)
	{
		*ppVS = new VideoSourceQT(this, Track[i]);
		if (*ppVS != NULL)
		{
			(*ppVS)->AddRef();
			return true;
		}
	}
	else
	{
		// No video; find an audio track and generate dummy video

		for (i = 0; i < num_tracks; ++i)
		{
			if (Track[i] != NULL
				&& Track[i]->fccType == streamtypeAUDIO
				&& (Track[i]->dwFlags & QTT_UNSUPPORTED) == 0)
			{
				if (index == 0) break;
				--index;
			}
		}

		if (i < num_tracks)
		{
			*ppVS = new VideoSourceDummy(this, Track[i]);
			if (*ppVS != NULL)
			{
				(*ppVS)->AddRef();
				return true;
			}
		}
	}

	return false;
}


bool InputFileQT::GetAudioSource(int index, IVDXAudioSource **ppAS)
{
	int i;
	*ppAS = NULL;

	for (i = 0; i < num_tracks; ++i)
	{
		if (Track[i] != NULL
			&& Track[i]->fccType == streamtypeAUDIO
			&& (Track[i]->dwFlags & QTT_UNSUPPORTED) == 0)
		{
			if (index == 0) break;
			--index;
		}
	}

	if (i < num_tracks)
	{
		switch (Track[i]->dwFormat)
		{
			case A__MP3:
			case A_AC_3:
			case A_MP4A:
			case A_MS11:
			case A_MS_U:
			case A_SAMR:
			case A_SAWB:
				*ppAS = new AudioSourceQT(this, Track[i]);
				break;
			case A_IMA4:
				*ppAS = new AudioSource_ima4(this, Track[i]);
				break;
			case A_ALAW:
			case A_FL32:
			case A_FL64:
			case A_IN24:
			case A_IN32:
			case A_RAW_:
			case A_SOWT:
			case A_TWOS:
			case A_ULAW:
				*ppAS = new AudioSource_PCM(this, Track[i]);
				break;
		}

		if (*ppAS != NULL)
		{
			(*ppAS)->AddRef();
			return true;
		}
	}

	return false;
}


///////////////////////////////////////////////////////////////////////
//
//						InputFileDriverQT
//
///////////////////////////////////////////////////////////////////////


class InputFileDriverQT : public vdxunknown<IVDXInputFileDriver> {

public:
	InputFileDriverQT(const VDXInputDriverContext& context);

	int		VDXAPIENTRY DetectBySignature(const void *pHeader, sint32 nHeaderSize, const void *pFooter, sint32 nFooterSize, sint64 nFileSize);
	bool	VDXAPIENTRY CreateInputFile(uint32 flags, IVDXInputFile **ppFile);

protected:
	const VDXInputDriverContext& mContext;
};


InputFileDriverQT::InputFileDriverQT(const VDXInputDriverContext& context)
	: mContext(context)
{
}


int VDXAPIENTRY InputFileDriverQT::DetectBySignature(const void *pHeader, sint32 nHeaderSize, const void *pFooter, sint32 nFooterSize, sint64 nFileSize)
{
	if (pHeader != NULL && nHeaderSize >= 16)
	{
		__int64 sz = *(unsigned long *)pHeader;
		unsigned long t = *((unsigned long *)pHeader + 1);
		bswap32(sz);

		if (sz == 0)
		{
			sz = nFileSize;
		}
		else if (sz == 1)
		{
			sz = *((__int64 *)pHeader + 1);
			bswap64(sz);
			if (sz < 16) goto Abort;
		}
		else
		{
			if (sz < 8) goto Abort;
		}

		if (sz <= nFileSize || nFileSize == 0)
		{
			switch (t) {
				case 'ediw':
				case 'pytf':
				case 'piks':
				case 'tadm':
				case 'voom':
					return 1;
			}
		}
	}

Abort:
	return -1;
}


bool VDXAPIENTRY InputFileDriverQT::CreateInputFile(uint32 flags, IVDXInputFile **ppFile)
{
	*ppFile = new InputFileQT(mContext);
	if (*ppFile == NULL) return false;

	(*ppFile)->AddRef();
	return true;
}


///////////////////////////////////////////////////////////////////////


bool VDXAPIENTRY QT_create(const VDXInputDriverContext *pContext, IVDXInputFileDriver **ppDriver)
{
	*ppDriver = new InputFileDriverQT(*pContext);
	if (*ppDriver == NULL) return false;

	(*ppDriver)->AddRef();
	return true;
}


const VDXInputDriverDefinition QT_input = {
	sizeof(VDXInputDriverDefinition),
	VDXInputDriverDefinition::kFlagSupportsVideo
		| VDXInputDriverDefinition::kFlagSupportsAudio
		| VDXInputDriverDefinition::kFlagCustomSignature,
	0,
	0,
	NULL,
	L"*.mov|*.mp4|*.m4v|*.m4a|*.qt|*.3gp|*.3g2|*.f4v",
	L"QuickTime files ("
	L"*.mov,*.mp4,*.m4v,*.m4a,*.qt,*.3gp,*.3g2,*.f4v)|"
	L"*.mov;*.mp4;*.m4v;*.m4a;*.qt;*.3gp;*.3g2;*.f4v",
	L"QuickTime",
	QT_create
};


// (major<<24) + (minor<<16) + build.  1.4.1000 would be 0x010403E8.

//	xxxxxxxx ........ ........ ........		major
//	........ xxxxxxxx ........ ........		minor
//	........ ........ xxxxxxxx xxxxxxxx		build

const VDXPluginInfo QT_info = {
	sizeof(VDXPluginInfo),              // size
	L"QuickTime plugin",                // name
	L"fccHandler",                      // author
	L"Loads QuickTime files.",          // description
	0x02060000,                         // (major<<24) + (minor<<16) + build
	kVDXPluginType_Input,               // type
	1,                                  // flags
	10,									// version required
	kVDXPlugin_APIVersion,              // version used
	2,									// version required
	kVDXPlugin_InputDriverAPIVersion,   // version used
	&QT_input
};


///////////////////////////////////////////////////////////////////////

const VDXPluginInfo *const kPlugins[] = {
	&QT_info,
	NULL
};


extern "C" const VDXPluginInfo *const * VDXAPIENTRY VDGetPluginInfo()
{
	return kPlugins;
}

