///////////////////////////////////////////////////////////////////////
// QuickTime plugin for VirtualDub 1.8.8+
// Copyright (C) 2007-2012 fccHandler
// Copyright (C) 1998-2012 Avery Lee
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
///////////////////////////////////////////////////////////////////////

// QuickTime Track class

#define WIN32_LEAN_AND_MEAN
#include <windows.h>
#include <vfw.h>

#include "QTTypes.h"
#include "QTTrack.h"

QTTrack::QTTrack()
	: dwTrackID             (0)

	, dwTimeScale           (0)
	, i64Duration           (0)
	, dwRate                (0)
	, dwScale               (0)

	, fccType               (0)

	, dwFormat              (0)
	, pFormat               (NULL)
	, lFormatLen            (0)
	, dwSamplesPerPacket    (0)
	, dwBytesPerPacket      (0)
	, dwBytesPerFrame       (0)
	, dwBytesPerSample      (0)

	, iEndianness           (0)

	, hSpacing              (0)
	, vSpacing              (0)

	, chunk_info            (NULL)
	, num_chunk_infos       (0)

	, chunk_offset32        (NULL)
	, chunk_offset64        (NULL)
	, num_chunk_offsets     (0)

	, keyframe              (NULL)
	, num_keyframes         (0)

	, time_to_sample_table  (NULL)
	, time_to_sample_count  (0)
	, sample_map            (NULL)

	, sample_size           (NULL)
	, default_sample_size   (0)
	, num_samples           (0)
	, perceived_samples     (0)
	, sample_list           (NULL)

	, dwFlags               (0)
	, dwMaxSample           (0)
	, i64TotalBytes         (0)

	, avc_dcr               (NULL)
	, dcr_size              (0)
	, avc_len               (0)

	, i64bytepos            (0)
	, dwCurBlock            (0)
{
}


QTTrack::~QTTrack()
{
	delete[] sample_map;
	delete[] sample_list;
	delete[] pFormat;
	delete[] avc_dcr;
}


bool QTTrack::Verify(const QTTrack &qtt)
{
	// Check the validity of a QTTrack.
	// (This is a static method which all Tracks share.)

	// 'mdia' atom
	if (qtt.dwTimeScale == 0) goto Abort;
	if (qtt.i64Duration == 0) goto Abort;
	// dwRate and dwScale are determined later.

	// 'hdlr' atom	
	if (qtt.fccType != streamtypeVIDEO)
	{
		if (qtt.fccType != streamtypeAUDIO) goto Abort;
	}

	// 'stsd' atom
	// dwFormat is checked elsewhere
	if (qtt.pFormat == NULL) goto Abort;
	if (qtt.lFormatLen <= 0) goto Abort;
	if (qtt.fccType == streamtypeVIDEO)
	{
		const BITMAPINFOHEADER* bmih = (BITMAPINFOHEADER *)qtt.pFormat;
		if (bmih->biWidth  <= 0 || bmih->biWidth  > MAX_VIDEO_WIDTH ) goto Abort;
		if (bmih->biHeight <= 0 || bmih->biHeight > MAX_VIDEO_HEIGHT) goto Abort;
	}
	// SoundDescriptionV1 is optional

	// 'stsc' atom
	if (qtt.chunk_info == NULL) goto Abort;
	if (qtt.num_chunk_infos == 0) goto Abort;

	// 'stco' atom
	if (qtt.chunk_offset32 == NULL)
	{
		if (qtt.chunk_offset64 == NULL) goto Abort;
	}
	if (qtt.num_chunk_offsets == 0) goto Abort;

	// 'stss' atom is optional
	
	// 'stts' atom
	if (qtt.time_to_sample_count == 0) goto Abort;
	if (qtt.time_to_sample_table == NULL) goto Abort;

	// 'stsz' atom
	if (qtt.default_sample_size == 0)
	{
		if (qtt.sample_size == NULL) goto Abort;
	}
	if (qtt.num_samples == 0) goto Abort;

	// perceived_samples and sample_list are derived later.

	// QTT_HAS_VSB, dwMaxSample, and i64TotalBytes are derived later.

	// AVCDecoderConfigurationRecord, or MPEG-4 Pt.2 VOL
	if (qtt.dwFormat == V_AVC1 /*|| qtt.dwFormat == V_MP4V*/)
	{
		if (qtt.avc_dcr == NULL) goto Abort;
		if (qtt.dcr_size <= 0) goto Abort;

		if (qtt.dwFormat == V_AVC1)
		{
			if (qtt.avc_len < 1 || qtt.avc_len > 4)
			{
				goto Abort;
			}
		}
	}

	// TODO: Anything else we need to verify?

	return true;

Abort:
	return false;
}


void QTTrack::DeriveFrameRate()
{
	// Determine the video frame rate for presentation.

	unsigned long i, dur;

	dwRate  = dwTimeScale;
	dwScale = time_to_sample_table[0].duration;

	// Some questionable movies have durations of zero,
	// so find the first non-zero duration.

	for (i = 0; (i + 1) < time_to_sample_count; ++i)
	{
		dwScale = time_to_sample_table[i].duration;
		if (dwScale != 0) break;
	}

	// The duration of the last sample is often much less
	// than the duration of the others, therefore skip it.

	for (i = 0; (i + 1) < time_to_sample_count; ++i)
	{
		dur = time_to_sample_table[i].duration;
		if (dur != 0 && dur < dwScale) dwScale = dur;
	}
	if (time_to_sample_table[i].count > 1)
	{
		dur = time_to_sample_table[i].duration;
		if (dur != 0 && dur < dwScale) dwScale = dur;
	}

	// Explanation:
	// dwDuration / least = perceived samples
	// perceived_samples / dur = maximum rate
	// Some real-world examples follow:
	
	// Example: Rosmeri Wind.mp4
	//     dwDuration = 31280968
	//     dwTimeScale = 90000
	//     num_samples = 10427
	//     least = 2999
	// duration = 31280968 / 90000 = 347.56631 secs
	// perceived_samples = 31280968 / 2999 = 10430.466
	// maximum_rate = 10430.466 / (31280968 / 90000) = 30.010002 fps

	// Example: armadillomd.mov
	//     dwDuration = 6368000
	//     dwTimeScale = 29970
	//     num_samples = 6368
	//     least = 1000
	// duration = 6368000 / 29970 = 212.47914 secs
	// perceived_samples = 6368000 / 1000 = 6368
	// maximum_rate = 6368 / (6368000 / 29970) = 29.97 fps

	// Example: TZ107 013163 Mute.mp4
	//     dwDuration = 275858582
	//     dwTimeScale = 90000
	//     num_samples = 75126
	//     least = 3003
	// duration = 275858582 / 90000 = 3065.09536 secs
	// perceived_samples = 275858582 / 3003 = 91860.9997
	// maximum_rate = 91860.9997 / (275858582 / 90000) = 29.97 fps

	// Unfortunately I've see a few like the following:
	// Example: Sade Danielle.mp4
	//     dwDuration = 267809
	//     dwTimeScale = 1000
	//     num_samples = 8026
	//     least = 5 (!)
	// duration = 267809 / 1000 = 267.809 secs
	// perceived_samples = 267809 / 5 = 53561.8
	// maximum_rate = 53561.8 / (267809 / 1000) = 200 fps (!)
	// The overwhelming number of frames differ by 33 or 34 ms,
	// indicating an intended rate of 29.97 fps.  However, the
	// presence of one 5 ms difference in the list means that
	// to display that frame at its proper time we must go to
	// 200 fps.  I don't have a good workaround yet, aargh...

	// Break down the formula:
	// (a / b) / c = a / (b * c)
	// a / (b / c) = (a * c) / b

	// max = (dwDuration / least) / (dwDuration / dwTimeScale)
	// max = ((dwDuration / least) * dwTimeScale) / dwDuration
	// max = ((dwDuration * dwTimeScale) / least) / dwDuration
	// max = (dwDuration * dwTimeScale) / (least * dwDuration)
	// max = dwTimeScale / least !!!
}


void QTTrack::DeriveSampleMap()
{
	// Generate a mapping of perceived frame to real sample.
	// Only call this after dwRate and dwScale are known!
	// The output is "perceived_samples" and "sample_map[]"

	DWORD dwEntry;
	DWORD dwCount;
	DWORD dwSampleDur;
	DWORD dwSampleIn;
	DWORD dwSampleOut;

	__int64 iSampleTime;
	__int64 iPerceivedTime;

	perceived_samples = (unsigned long)((i64Duration + (dwScale - 1)) / dwScale);

	sample_map = new unsigned long[perceived_samples];
	if (sample_map == NULL) goto Abort;

	dwSampleIn     = 0;
	dwSampleOut    = 0;
	iSampleTime    = 0;
	iPerceivedTime = 0;

	// Given: num_samples is not zero

	for (dwEntry = 0; dwEntry < time_to_sample_count; ++dwEntry)
	{
		dwSampleDur = time_to_sample_table[dwEntry].duration;
		for (dwCount = 0; dwCount < time_to_sample_table[dwEntry].count; ++dwCount)
		{
			iSampleTime += dwSampleDur;
			// Catch-up perceived time
			while (iPerceivedTime < iSampleTime)
			{
				if (dwSampleOut >= perceived_samples) goto Abort;
				sample_map[dwSampleOut++] = dwSampleIn;
				iPerceivedTime += dwScale;
			}
			++dwSampleIn;
			if (dwSampleIn >= num_samples)
			{
				dwEntry = time_to_sample_count;
				break;
			}
		}
	}

	// Finish if necessary
	while (iPerceivedTime < i64Duration)
	{
		if (dwSampleOut >= perceived_samples) goto Abort;
		sample_map[dwSampleOut++] = num_samples - 1;
		iPerceivedTime += dwScale;
	}

Abort:
	return;
}

/*
void QTTrack::DeriveTimeStamps()
{
	// Derive sample time stamps (in milliseconds).

	DWORD dwEntry;
	DWORD dwCount;
	DWORD dwSampleDur;
	DWORD dwSampleOut;

	__int64 iSampleTime;

	dwSampleOut = 0;
	iSampleTime = 0;

	for (dwEntry = 0; dwEntry < time_to_sample_count; ++dwEntry)
	{
		dwSampleDur = time_to_sample_table[dwEntry].duration;
		for (dwCount = 0; dwCount < time_to_sample_table[dwEntry].count; ++dwCount)
		{
			if (dwSampleOut >= num_samples) break;

			sample_list[dwSampleOut++].pts = (unsigned long)
				((1000.0 * (double)iSampleTime) / (double)dwTimeScale + 0.5);

			iSampleTime += dwSampleDur;
		}
	}
}
*/

bool QTTrack::Finish()
{
	// The default implementation of Finish is suitable
	// for video tracks only! For audio tracks it is
	// necessary to override the method in a derived class.

	// Derive sample positions using the chunk tables

	// Example:
	//           +- Sample 1
	// Chunk 1 --|  Sample 2
	//           +- Sample 3
	//           +- Sample 4
	// Chunk 2 --|  Sample 5
	//           +- Sample 6
	// Chunk 3 --+- Sample 7
	// Chunk 4 --+- Sample 8
	// Chunk 5 --+- Sample 9

	// Sample To Chunk table:
	// First Chunk: 1  Samples Per Chunk: 3  Sample Description ID: 23
	// First Chunk: 3  Samples Per Chunk: 1  Sample Description ID: 23
	// First Chunk: 5  Samples Per Chunk: 1  Sample Description ID: 24

	unsigned long ci;
	unsigned long co;
	unsigned long limit;

	unsigned long sample;
	unsigned long s;
	__int64 pos;

	bool ret = false;

	// Be aware of what has already been Verified (see above).

	if (fccType != streamtypeVIDEO)
	{
		// Audio track got here because we don't have
		// a handler for it!
		num_samples = 0;
		dwFlags |= QTT_UNSUPPORTED;
		return true;
	}

	perceived_samples = num_samples;
	sample_list = new QTSampleInfo[num_samples];
	if (sample_list == NULL) goto Abort;

	for (ci = 0; ci < num_samples; ++ci)
	{
		// Note: zero is not a valid position
		sample_list[ci].pos  = 0;
		sample_list[ci].size = 0;
	}

	// Derive sample sizes

	if (default_sample_size != 0)
	{
		for (ci = 0; ci < num_samples; ++ci)
		{
			sample_list[ci].size = default_sample_size;
		}
	}
	else
	{
		for (ci = 0; ci < num_samples; ++ci)
		{
			sample_list[ci].size = sample_size[ci];
		}
	}

	// Derive sample positions

	sample = 0;

	for (ci = 0; ci < num_chunk_infos; ++ci)
	{
		if ((ci + 1) < num_chunk_infos)
		{
			limit = chunk_info[ci + 1].first;
		}
		else
		{
			limit = num_chunk_offsets + 1;
		}

		for (co = chunk_info[ci].first; co < limit; ++co)
		{
			if (co < 1 || co > num_chunk_offsets) goto Abort;

			if (chunk_offset32 == NULL)
			{
				pos = chunk_offset64[co - 1];
			}
			else
			{
				pos = chunk_offset32[co - 1];
			}
				
			for (s = 0; s < chunk_info[ci].samples; ++s)
			{
				if (sample >= num_samples) goto Abort;
				sample_list[sample].pos = pos;
				pos += sample_list[sample].size;
				++sample;
			}
		}
	}

	// Derive total bytes and maximum sample size

	dwMaxSample   = sample_list[0].size;
	i64TotalBytes = dwMaxSample;

	for (ci = 1; ci < num_samples; ++ci)
	{
		co = sample_list[ci].size;
		if (co != dwMaxSample)
		{
			dwFlags |= QTT_HAS_VSB;
			if (co > dwMaxSample) dwMaxSample = co;
		}
		i64TotalBytes += co;
	}
	
	// Mark the keyframes

	if (keyframe != NULL)
	{
		for (ci = 0; ci < num_keyframes; ++ci)
		{
			co = keyframe[ci];
			if (co >= 1 && co <= num_samples)
			{
				sample_list[co - 1].size |= QT_KEYFRAME;
			}
		}
	}
	else
	{
		// All frames are keyframes
		for (ci = 0; ci < num_samples; ++ci)
		{
			sample_list[ci].size |= QT_KEYFRAME;
		}
	}

//	DeriveTimeStamps();

	DeriveFrameRate();
	if (dwRate  == 0) goto Abort;
	if (dwScale == 0) goto Abort;

	DeriveSampleMap();
	if (sample_map == NULL) goto Abort;
	if (perceived_samples == 0) goto Abort;

	ret = true;		// All OK

Abort:
	return ret;
}


/***************************************************************
 *          Dolby Digital AC-3
 ***************************************************************/

bool QTTrack_ac_3::Finish()
{
	WAVEFORMATEX *wfex;

	unsigned long ci;
	unsigned long co;
	unsigned long limit;

	unsigned long sample;
	unsigned long s;
	__int64 pos;

	bool ret = false;

	// Be aware of what has already been Verified (see above).

	wfex = (WAVEFORMATEX *)pFormat;
	wfex->wFormatTag = 0x2000;

	perceived_samples = num_samples;
	sample_list = new QTSampleInfo[num_samples];
	if (sample_list == NULL) goto Abort;

	for (ci = 0; ci < num_samples; ++ci)
	{
		sample_list[ci].pos = 0;
	}

	if (default_sample_size != 0)
	{
		for (ci = 0; ci < num_samples; ++ci)
		{
			sample_list[ci].size = default_sample_size;
		}
	}
	else
	{
		for (ci = 0; ci < num_samples; ++ci)
		{
			sample_list[ci].size = sample_size[ci];
		}
	}

	sample = 0;

	for (ci = 0; ci < num_chunk_infos; ++ci)
	{
		if ((ci + 1) < num_chunk_infos)
		{
			limit = chunk_info[ci + 1].first;
		}
		else
		{
			limit = num_chunk_offsets + 1;
		}

		for (co = chunk_info[ci].first; co < limit; ++co)
		{
			if (co < 1 || co > num_chunk_offsets) goto Abort;

			if (chunk_offset32 == NULL)
			{
				pos = chunk_offset64[co - 1];
			}
			else
			{
				pos = chunk_offset32[co - 1];
			}
			
			for (s = 0; s < chunk_info[ci].samples; ++s)
			{
				if (sample >= num_samples) goto Abort;
				sample_list[sample].pos = pos;
				pos += sample_list[sample].size;
				++sample;
			}
		}
	}

	if (dwSamplesPerPacket == 0) dwSamplesPerPacket = 1536;

	// Calculate total bytes and maximum sample size

	dwMaxSample   = sample_list[0].size;
	i64TotalBytes = dwMaxSample;

	for (ci = 1; ci < num_samples; ++ci)
	{
		co = sample_list[ci].size;
		i64TotalBytes += co;

		if (co != dwMaxSample)
		{
			dwFlags |= QTT_HAS_VSB;
			if (co > dwMaxSample) dwMaxSample = co;
		}
	}
	
	wfex->nAvgBytesPerSec = (DWORD)(
		((double)i64TotalBytes * (double)wfex->nSamplesPerSec) /
		((double)dwSamplesPerPacket * (double)num_samples) + 0.5
	);
	
	wfex->nBlockAlign = (WORD)dwMaxSample;
	wfex->wBitsPerSample = 0;

	// Sample rate
	dwRate  = wfex->nSamplesPerSec;
	dwScale = dwSamplesPerPacket;

	ret = true;		// All OK

Abort:
	return ret;
}


/***************************************************************
 *          Floating point PCM (little-endian)
 ***************************************************************/

bool QTTrack_float::Finish()
{
	WAVEFORMATEX *wfex;
	unsigned long ci;
	unsigned long co;
	unsigned long limit;

	bool ret = false;

	// Be aware of what has already been Verified (see above).

	if (default_sample_size == 0) goto Abort;

	wfex = (WAVEFORMATEX *)pFormat;

	if (wfex->nChannels != 1 && wfex->nChannels != 2) goto Abort;
	if (wfex->wBitsPerSample != 16) goto Abort;

	wfex->wBitsPerSample = (dwFormat == A_FL32)? 32: 64;

	// Block alignment is needed for parsing the chunks.
	wfex->nBlockAlign = wfex->nChannels * (wfex->wBitsPerSample / 8);

	perceived_samples = 0;
	num_samples = num_chunk_offsets;

	sample_list = new QTSampleInfo[num_samples];
	if (sample_list == NULL) goto Abort;

	for (ci = 0; ci < num_samples; ++ci)
	{
		// Note: zero is not a valid position
		sample_list[ci].pos  = 0;
		sample_list[ci].size = 0;
	}

	for (ci = 0; ci < num_chunk_infos; ++ci)
	{
		if ((ci + 1) < num_chunk_infos)
		{
			limit = chunk_info[ci + 1].first;
		}
		else
		{
			limit = num_chunk_offsets + 1;
		}

		for (co = chunk_info[ci].first; co < limit; ++co)
		{
			if (co < 1 || co > num_chunk_offsets) goto Abort;

			if (chunk_offset32 == NULL)
			{
				sample_list[co - 1].pos = chunk_offset64[co - 1];
			}
			else
			{
				sample_list[co - 1].pos = chunk_offset32[co - 1];
			}

			sample_list[co - 1].size = chunk_info[ci].samples * wfex->nBlockAlign;

			perceived_samples += chunk_info[ci].samples;
		}
	}

	// Calculate total bytes and maximum sample size

	dwMaxSample   = sample_list[0].size;
	i64TotalBytes = dwMaxSample;

	for (ci = 1; ci < num_samples; ++ci)
	{
		co = sample_list[ci].size;
		if (co > dwMaxSample) dwMaxSample = co;
		i64TotalBytes += co;
	}

	wfex->wFormatTag      = WAVE_FORMAT_IEEE_FLOAT;
	wfex->nAvgBytesPerSec = wfex->nBlockAlign * wfex->nSamplesPerSec;
	wfex->cbSize          = 0;

	dwRate  = wfex->nSamplesPerSec * wfex->nBlockAlign;
	dwScale = wfex->nBlockAlign;

	ret = true;		// All OK

Abort:
	return ret;
}


/***************************************************************
 *          MPEG Layer III (CBR / VBR)
 ***************************************************************/

bool QTTrack__mp3::Finish()
{
	MPEGLAYER3WAVEFORMAT *wf;

	unsigned long ci;
	unsigned long co;
	unsigned long limit;

	unsigned long sample;
	unsigned long s;
	__int64 pos;

	bool ret = false;

	// Be aware of what has already been Verified (see above).

	if (default_sample_size != 0) goto Abort;

	perceived_samples = num_samples;
	sample_list = new QTSampleInfo[num_samples];
	if (sample_list == NULL) goto Abort;

	// Derive sample sizes
	for (ci = 0; ci < num_samples; ++ci)
	{
		sample_list[ci].pos  = 0;
		sample_list[ci].size = sample_size[ci];
	}

	sample = 0;

	for (ci = 0; ci < num_chunk_infos; ++ci)
	{
		if ((ci + 1) < num_chunk_infos)
		{
			limit = chunk_info[ci + 1].first;
		}
		else
		{
			limit = num_chunk_offsets + 1;
		}

		for (co = chunk_info[ci].first; co < limit; ++co)
		{
			if (co < 1 || co > num_chunk_offsets) goto Abort;

			if (chunk_offset32 == NULL)
			{
				pos = chunk_offset64[co - 1];
			}
			else
			{
				pos = chunk_offset32[co - 1];
			}
				
			for (s = 0; s < chunk_info[ci].samples; ++s)
			{
				if (sample >= num_samples) goto Abort;
				sample_list[sample].pos = pos;
				pos += sample_list[sample].size;
				++sample;
			}
		}
	}

	// Calculate total bytes and maximum sample size

	dwMaxSample   = sample_list[0].size;
	i64TotalBytes = dwMaxSample;

	for (ci = 1; ci < num_samples; ++ci)
	{
		co = sample_list[ci].size;
		i64TotalBytes += co;

		if (co != dwMaxSample)
		{
			dwFlags |= QTT_HAS_VSB;
			if (co > dwMaxSample) dwMaxSample = co;
		}
	}

	if (dwSamplesPerPacket == 0) dwSamplesPerPacket = 1152;

	wf = (MPEGLAYER3WAVEFORMAT *)pFormat;

	wf->wfx.nAvgBytesPerSec = (DWORD)(
		((double)i64TotalBytes * (double)wf->wfx.nSamplesPerSec) /
		((double)dwSamplesPerPacket * (double)num_samples) + 0.5
	);

	wf->wfx.wFormatTag     = WAVE_FORMAT_MPEGLAYER3;
	wf->wfx.nBlockAlign    = (WORD)dwMaxSample;
	wf->wfx.wBitsPerSample = 0;
	wf->wfx.cbSize         = MPEGLAYER3_WFX_EXTRA_BYTES;
	wf->wID                = MPEGLAYER3_ID_MPEG;
	wf->fdwFlags           = MPEGLAYER3_FLAG_PADDING_ISO;
	wf->nBlockSize         = (WORD)dwMaxSample;
	wf->nFramesPerBlock    = 1;
	wf->nCodecDelay        = 0;

	dwRate  = wf->wfx.nSamplesPerSec;
	dwScale = dwSamplesPerPacket;

	ret = true;		// All OK

Abort:
	return ret;
}


/***************************************************************
 *          QuickTime ADPCM 4-bit
 ***************************************************************/

bool QTTrack_ima4::Finish()
{
	WAVEFORMATEX *wfex;

	unsigned long ci;
	unsigned long co;
	unsigned long limit;

	bool ret = false;

	// Be aware of what has already been Verified (see above).

	if (default_sample_size == 0) goto Abort;

	if (dwSamplesPerPacket == 0)
	{
		dwSamplesPerPacket = 64; // default
	}
	else
	{
		if (dwSamplesPerPacket & 1) goto Abort;
	}

	wfex = (WAVEFORMATEX *)pFormat;

	wfex->wFormatTag      = WAVE_FORMAT_PCM;
	wfex->nBlockAlign     = wfex->nChannels * (wfex->wBitsPerSample / 8);
	wfex->nAvgBytesPerSec = wfex->nBlockAlign * wfex->nSamplesPerSec;
	wfex->cbSize          = 0;

	perceived_samples = 0;
	num_samples = num_chunk_offsets;

	sample_list = new QTSampleInfo[num_samples];
	if (sample_list == NULL) goto Abort;

	for (ci = 0; ci < num_samples; ++ci)
	{
		sample_list[ci].pos  = 0;
		sample_list[ci].size = 0;
	}

	for (ci = 0; ci < num_chunk_infos; ++ci)
	{
		if ((ci + 1) < num_chunk_infos)
		{
			limit = chunk_info[ci + 1].first;
		}
		else
		{
			limit = num_chunk_offsets + 1;
		}

		for (co = chunk_info[ci].first; co < limit; ++co)
		{
			if (co < 1 || co > num_chunk_offsets) goto Abort;

			if (chunk_offset32 == NULL)
			{
				sample_list[co - 1].pos = chunk_offset64[co - 1];
			}
			else
			{
				sample_list[co - 1].pos = chunk_offset32[co - 1];
			}

			sample_list[co - 1].size =
				(chunk_info[ci].samples / dwSamplesPerPacket)
				* (2 + (dwSamplesPerPacket >> 1)) * wfex->nChannels;

			perceived_samples += chunk_info[ci].samples;
		}
	}

	// Calculate total bytes and maximum sample size

	dwMaxSample   = sample_list[0].size;
	i64TotalBytes = dwMaxSample;

	for (ci = 1; (unsigned long)ci < num_samples; ++ci)
	{
		co = sample_list[ci].size;
		if (co > dwMaxSample) dwMaxSample = co;
		i64TotalBytes += co;
	}
	
	dwRate  = wfex->nSamplesPerSec * wfex->nBlockAlign;
	dwScale = wfex->nBlockAlign;
	
	ret = true;		// All OK

Abort:
	return ret;
}


/***************************************************************
 *          Advanced Audio Codec
 ***************************************************************/

bool QTTrack_mp4a::Finish()
{
	static const unsigned char chan_map[16] = {
		0, 1, 2, 3, 4, 5, 6, 8, 0, 0, 0, 0, 0, 0, 0, 0
	};

	static const unsigned long freq_map[16] = {
		96000, 88200, 64000, 48000, 44100, 32000, 24000, 22050,
		16000, 12000, 11025,  8000,  7350,     0,     0,     0
	};

	WAVEFORMATEX *wfex;

	unsigned long ci;
	unsigned long co;
	unsigned long limit;
	unsigned long adts;

	unsigned long sample;
	unsigned long s;
	__int64 pos;

	bool ret = false;

	// Be aware of what has already been Verified (see above).

	wfex = (WAVEFORMATEX *)pFormat;

	if (wfex->cbSize < 2) goto Abort;

	s = *(WORD *)(pFormat + sizeof(WAVEFORMATEX));
	if (s != 0)
	{
//		if (default_sample_size != 0) goto Abort;

		wfex->wFormatTag = WAVE_FORMAT_RAW_AAC1;

		// InputFileQT sets the following flag if we are
		// wrapping AAC frames with ADTS headers:
		adts = (dwFlags & QTT_ADD_ADTS)? 7: 0;

		// AudioSpecificConfig:
		//   LSB      MSB
		// xxxxx... ........  object type (00010 = AAC LC)
		// .....xxx x.......  sampling rate
		// ........ .xxxx...  channel configuration
		// ........ .....x..  frame length flag (0 = 1024, 1 = 960)
		// ........ ......x.  depends on core coder (0)
		// ........ .......x  extension flag (0)

		bswap16(s);

		wfex->nSamplesPerSec = freq_map[(s >> 7) & 15];
		if (wfex->nSamplesPerSec == 0) goto Abort;

		ci = (s >> 3) & 15;
		if (ci == 0)
		{
			// Hack! birthdaygirl.mp4
			// For now just assume Parametric Stereo.
			if (wfex->cbSize <= 2) goto Abort;
			wfex->nChannels = 2;
		}
		else
		{
			wfex->nChannels = chan_map[ci];
			if (wfex->nChannels == 0) goto Abort;
		}

		if (dwSamplesPerPacket == 0)
		{
			dwSamplesPerPacket = (s & 4)? 960: 1024;
		}

		// Implicit SBR?
		if (wfex->nSamplesPerSec <= 24000)
		{
			wfex->nSamplesPerSec *= 2;
			dwSamplesPerPacket *= 2;
		}
	}
	else
	{
		// MPEG audio, hopefully layer III

		wfex->wFormatTag = WAVE_FORMAT_MPEGLAYER3;
		adts = 0;

		if (dwSamplesPerPacket == 0) dwSamplesPerPacket = 1152;
	}

	perceived_samples = num_samples;
	sample_list = new QTSampleInfo[num_samples];
	if (sample_list == NULL) goto Abort;

	for (ci = 0; ci < num_samples; ++ci)
	{
		sample_list[ci].pos = 0;
	}

	if (default_sample_size != 0)
	{
		for (ci = 0; ci < num_samples; ++ci)
		{
			sample_list[ci].size = default_sample_size;
		}
	}
	else
	{
		for (ci = 0; ci < num_samples; ++ci)
		{
			sample_list[ci].size = sample_size[ci];
		}
	}

	sample = 0;

	for (ci = 0; ci < num_chunk_infos; ++ci)
	{
		if ((ci + 1) < num_chunk_infos)
		{
			limit = chunk_info[ci + 1].first;
		}
		else
		{
			limit = num_chunk_offsets + 1;
		}

		for (co = chunk_info[ci].first; co < limit; ++co)
		{
			if (co < 1 || co > num_chunk_offsets) goto Abort;

			if (chunk_offset32 == NULL)
			{
				pos = chunk_offset64[co - 1];
			}
			else
			{
				pos = chunk_offset32[co - 1];
			}
			
			for (s = 0; s < chunk_info[ci].samples; ++s)
			{
				if (sample >= num_samples) goto Abort;
				sample_list[sample].pos = pos;
				pos += sample_list[sample].size;
				++sample;
			}
		}
	}

	// Calculate total bytes and maximum sample size

	dwMaxSample   = sample_list[0].size + adts;
	i64TotalBytes = dwMaxSample;

	for (ci = 1; ci < num_samples; ++ci)
	{
		co = sample_list[ci].size + adts;
		i64TotalBytes += co;

		if (co != dwMaxSample)
		{
			dwFlags |= QTT_HAS_VSB;
			if (co > dwMaxSample) dwMaxSample = co;
		}
	}

	// Average bytes per second:
	// avg = total_bytes / duration_in_seconds
	// avg = total_bytes / ((samples_per_packet * num_samples) / sample_rate)
	// avg = (total_bytes * sample_rate) / (samples_per_packet * num_samples)
	
	wfex->nAvgBytesPerSec = (DWORD)(
		((double)i64TotalBytes * (double)wfex->nSamplesPerSec) /
		((double)dwSamplesPerPacket * (double)num_samples) + 0.5
	);

	wfex->nBlockAlign = (WORD)dwMaxSample;
	wfex->wBitsPerSample = 0;

	dwRate  = wfex->nSamplesPerSec;
	dwScale = dwSamplesPerPacket;

	if (wfex->wFormatTag == WAVE_FORMAT_MPEGLAYER3)
	{
		MPEGLAYER3WAVEFORMAT *wf = (MPEGLAYER3WAVEFORMAT *)pFormat;
			
		wf->wfx.cbSize          = MPEGLAYER3_WFX_EXTRA_BYTES;
		wf->wID                 = MPEGLAYER3_ID_MPEG;
		wf->fdwFlags            = MPEGLAYER3_FLAG_PADDING_ISO;
		wf->nBlockSize          = (WORD)dwMaxSample;
		wf->nFramesPerBlock     = 1;
		wf->nCodecDelay         = 0;
	}

	ret = true;		// All OK

Abort:
	return ret;
}


void QTTrack_mp4a::ParseHeader(unsigned int hdr)
{
	// Adjust the format based on MPEG Audio header, see:
	// http://msdn.microsoft.com/en-us/library/windows/desktop/dd390710(v=vs.85).aspx
	// http://msdn.microsoft.com/en-us/library/ms787304(v=vs.85).aspx

	int ver;
	int layer;

	bswap32(hdr);

	// MPEG Audio header:
	// 11111111 1111.... ........ ........   sync (all bits set)
	// ........ ....x... ........ ........   version (1=MPEG-1, 0=MPEG-2)
	// ........ .....xx. ........ ........   layer (00=reserved,11=I,10=II,01=III)
	// ........ .......x ........ ........   protected by CRC (0=yes, 1=no)
	// ........ ........ xxxx.... ........   bitrate_index (can't be 1111)
	// ........ ........ ....xx.. ........   sampling_freq (can't be 11)
	// ........ ........ ......x. ........   padding_bit (0=no, 1=yes)
	// ........ ........ .......x ........   private_bit (not used)
	// ........ ........ ........ xx......   (00=stereo,10=joint,01=dual,11=mono)
	// ........ ........ ........ ..xx....   mode_ext (only if joint stereo)
	// ........ ........ ........ ....x...   copyright_bit (0=no, 1=yes)
	// ........ ........ ........ .....x..   original_or_copy (0=original, 1=copy)
	// ........ ........ ........ ......xx   emphasis (00=none,01=reserved,10=50/15ms,11=CCIT J.17)

	// Validate
	if ((hdr & 0xFFF00000) != 0xFFF00000) goto Abort;
	if ((hdr & 0x00060000) == 0x00000000) goto Abort;
	if ((hdr & 0x0000F000) == 0x0000F000) goto Abort;
	if ((hdr & 0x00000C00) == 0x00000C00) goto Abort;

	ver = (hdr & 0x00080000)? 1: 2;
	layer = 4 - ((hdr >> 17) & 3);

	if (layer == 3)
	{
		// Adjust for MPEG-2 Audio
		if (ver == 2)
		{
			MPEGLAYER3WAVEFORMAT *wf = (MPEGLAYER3WAVEFORMAT *)pFormat;

			dwSamplesPerPacket >>= 1;
			dwScale = dwSamplesPerPacket;

			wf->wfx.nAvgBytesPerSec = (DWORD)(
				((double)i64TotalBytes * (double)wf->wfx.nSamplesPerSec) /
				((double)dwSamplesPerPacket * (double)num_samples) + 0.5
			);
		}
	}
	else
	{
		// Uh oh, it's not Layer III.  Reconstruct
		// the relevant portions of the wave format.

		static const unsigned short bitrate[2][3][16] = {
			{
				{ 0, 32, 64, 96, 128, 160, 192, 224, 256, 288, 320, 352, 384, 416, 448, 0 },	// MPEG-1 layer I
				{ 0, 32, 48, 56,  64,  80,  96, 112, 128, 160, 192, 224, 256, 320, 384, 0 },	// MPEG-1 layer II
				{ 0, 32, 40, 48,  56,  64,  80,  96, 112, 128, 160, 192, 224, 256, 320, 0 },	// MPEG-1 layer III
			},
			{
				{ 0, 32, 48, 56,  64,  80,  96, 112, 128, 144, 160, 176, 192, 224, 256, 0 },	// MPEG-2 layer I
				{ 0,  8, 16, 24,  32,  40,  48,  56,  64,  80,  96, 112, 128, 144, 160, 0 },	// MPEG-2 layer II
				{ 0,  8, 16, 24,  32,  40,  48,  56,  64,  80,  96, 112, 128, 144, 160, 0 },	// MPEG-2 layer III
			}
		};

		MPEG1WAVEFORMAT *wfex = (MPEG1WAVEFORMAT *)pFormat;

		lFormatLen = sizeof(MPEG1WAVEFORMAT);

		if (layer == 1) dwSamplesPerPacket = 384;

		if (ver == 2)
		{
			dwSamplesPerPacket >>= 1;
			dwScale = dwSamplesPerPacket;
		}

		wfex->wfx.wFormatTag     = WAVE_FORMAT_MPEG;
		wfex->wfx.wBitsPerSample = 0;
		wfex->wfx.cbSize         = sizeof(MPEG1WAVEFORMAT) - FIELD_OFFSET(MPEG1WAVEFORMAT, fwHeadLayer);

		wfex->wfx.nAvgBytesPerSec = (DWORD)(
			((double)i64TotalBytes * (double)wfex->wfx.nSamplesPerSec) /
			((double)dwSamplesPerPacket * (double)num_samples) + 0.5
		);

		wfex->fwHeadLayer    = (layer == 1)? ACM_MPEG_LAYER1: ACM_MPEG_LAYER2;
		wfex->dwHeadBitrate  = bitrate[ver - 1][layer - 1][(hdr >> 12) & 15] * 1000;

		if ((hdr & 0xC0) == 0xC0)
		{
			// Mono audio
			wfex->fwHeadMode = ACM_MPEG_SINGLECHANNEL;

			wfex->fwHeadModeExt = 0;
		}
		else
		{
			// Stereo mode may change dynamically, so set all bits.
			wfex->fwHeadMode = ACM_MPEG_STEREO
			                 | ACM_MPEG_DUALCHANNEL
			                 | ACM_MPEG_JOINTSTEREO;

			wfex->fwHeadModeExt = 15;
		}

		wfex->wHeadEmphasis = (hdr & 3) + 1;

		wfex->fwHeadFlags = (ver == 1)? ACM_MPEG_ID_MPEG1: 0;

		if (!(hdr & 0x10000)) wfex->fwHeadFlags |= ACM_MPEG_PROTECTIONBIT;
		if (hdr & 0x00000100) wfex->fwHeadFlags |= ACM_MPEG_PRIVATEBIT;
		if (hdr & 0x00000008) wfex->fwHeadFlags |= ACM_MPEG_COPYRIGHT;
		if (hdr & 0x00000004) wfex->fwHeadFlags |= ACM_MPEG_ORIGINALHOME;

		wfex->dwPTSLow  = 0;
		wfex->dwPTSHigh = 0;
	}

Abort:
	return;
}


/***************************************************************
 *          MPEG Layer III (CBR only)
 ***************************************************************/

bool QTTrack_ms_u::Finish()
{
	MPEGLAYER3WAVEFORMAT *wf;

	unsigned long ci;
	unsigned long co;
	unsigned long limit;

	unsigned long sample;
	unsigned long s;
	__int64 pos;

	bool ret = false;

	// Be aware of what has already been Verified (see above).

	if (default_sample_size != 1) goto Abort;

	// This one is tricky...

	if (dwSamplesPerPacket == 0) goto Abort;
	if (dwBytesPerFrame == 0) goto Abort;

	num_samples = 0;

	for (ci = 0; ci < num_chunk_infos; ++ci)
	{
		if ((ci + 1) < num_chunk_infos)
		{
			limit = chunk_info[ci + 1].first;
		}
		else
		{
			limit = num_chunk_offsets + 1;
		}

		for (co = chunk_info[ci].first; co < limit; ++co)
		{
			num_samples += (chunk_info[ci].samples / dwSamplesPerPacket);
		}
	}

	if (num_samples == 0) goto Abort;

	perceived_samples = num_samples;
	sample_list = new QTSampleInfo[num_samples];
	if (sample_list == NULL) goto Abort;

	for (ci = 0; ci < num_samples; ++ci)
	{
		sample_list[ci].pos  = 0;
		sample_list[ci].size = dwBytesPerFrame;
	}

	sample = 0;

	for (ci = 0; ci < num_chunk_infos; ++ci)
	{
		if ((ci + 1) < num_chunk_infos)
		{
			limit = chunk_info[ci + 1].first;
		}
		else
		{
			limit = num_chunk_offsets + 1;
		}

		for (co = chunk_info[ci].first; co < limit; ++co)
		{
			if (co < 1 || co > num_chunk_offsets) goto Abort;

			if (chunk_offset32 == NULL)
			{
				pos = chunk_offset64[co - 1];
			}
			else
			{
				pos = chunk_offset32[co - 1];
			}

			for (s = 0; s < (chunk_info[ci].samples / dwSamplesPerPacket); ++s)
			{
				if (sample >= num_samples) goto Abort;
				sample_list[sample].pos = pos;
				pos += dwBytesPerFrame;
				++sample;
			}
		}
	}

	// Calculate total bytes and maximum sample size

	dwMaxSample   = dwBytesPerFrame;
	i64TotalBytes = (__int64)dwBytesPerFrame * (__int64)num_samples;

	if (dwSamplesPerPacket == 0) dwSamplesPerPacket = 1152;

	wf = (MPEGLAYER3WAVEFORMAT *)pFormat;

	wf->wfx.nAvgBytesPerSec = (DWORD)(
		((double)i64TotalBytes * (double)wf->wfx.nSamplesPerSec) /
		((double)dwSamplesPerPacket * (double)num_samples) + 0.5
	);
			
	wf->wfx.wFormatTag      = WAVE_FORMAT_MPEGLAYER3;
	wf->wfx.nBlockAlign     = (WORD)dwMaxSample;
	wf->wfx.wBitsPerSample  = 0;
	wf->wfx.cbSize          = MPEGLAYER3_WFX_EXTRA_BYTES;
	wf->wID                 = MPEGLAYER3_ID_MPEG;
	wf->fdwFlags            = MPEGLAYER3_FLAG_PADDING_ISO;
	wf->nBlockSize          = (WORD)dwMaxSample;
	wf->nFramesPerBlock     = 1;
	wf->nCodecDelay         = 0;

	dwRate  = wf->wfx.nSamplesPerSec;
	dwScale = dwSamplesPerPacket;

	ret = true;		// All OK

Abort:
	return ret;
}


/***************************************************************
 *          Microsoft IMA ADPCM
 ***************************************************************/

bool QTTrack_ms11::Finish()
{
	WAVEFORMATEX *wfex;
	unsigned long ci;
	unsigned long co;
	unsigned long limit;

	bool ret = false;

	// Be aware of what has already been Verified (see above).

	if (default_sample_size == 0) goto Abort;
	if (dwSamplesPerPacket == 0) goto Abort;

	wfex = (WAVEFORMATEX *)pFormat;

	if (wfex->nChannels != 1 && wfex->nChannels != 2) goto Abort;
	if (wfex->wBitsPerSample != 4) goto Abort;
	if (wfex->nBlockAlign == 0) goto Abort;

	perceived_samples = 0;
	num_samples = num_chunk_offsets;

	sample_list = new QTSampleInfo[num_samples];
	if (sample_list == NULL) goto Abort;

	for (ci = 0; ci < num_samples; ++ci)
	{
		sample_list[ci].pos  = 0;
		sample_list[ci].size = 0;
	}

	for (ci = 0; ci < num_chunk_infos; ++ci)
	{
		if ((ci + 1) < num_chunk_infos)
		{
			limit = chunk_info[ci + 1].first;
		}
		else
		{
			limit = num_chunk_offsets + 1;
		}

		for (co = chunk_info[ci].first; co < limit; ++co)
		{
			if (co < 1 || co > num_chunk_offsets) goto Abort;

			if (chunk_offset32 == NULL)
			{
				sample_list[co - 1].pos = chunk_offset64[co - 1];
			}
			else
			{
				sample_list[co - 1].pos = chunk_offset32[co - 1];
			}

			sample_list[co - 1].size =
				(chunk_info[ci].samples / dwSamplesPerPacket)
				* wfex->nBlockAlign;

			perceived_samples += chunk_info[ci].samples;
		}
	}

	// Calculate total bytes and maximum sample size

	dwMaxSample   = sample_list[0].size;
	i64TotalBytes = dwMaxSample;

	for (ci = 1; ci < num_samples; ++ci)
	{
		co = sample_list[ci].size;
		if (co > dwMaxSample) dwMaxSample = co;
		i64TotalBytes += co;
	}

	dwRate  = wfex->nSamplesPerSec;
	dwScale = dwSamplesPerPacket;

	ret = true;		// All OK

Abort:
	return ret;
}


/***************************************************************
 *          Raw PCM (Intel or Motorola)
 ***************************************************************/

bool QTTrack_PCM::Finish()
{
	// Derive sample positions using the chunk tables

	// Example:
	//           +- Sample 1
	// Chunk 1 --|  Sample 2
	//           +- Sample 3
	//           +- Sample 4
	// Chunk 2 --|  Sample 5
	//           +- Sample 6
	// Chunk 3 --+- Sample 7
	// Chunk 4 --+- Sample 8
	// Chunk 5 --+- Sample 9

	// Sample To Chunk table:
	// First Chunk: 1  Samples Per Chunk: 3  Sample Description ID: 23
	// First Chunk: 3  Samples Per Chunk: 1  Sample Description ID: 23
	// First Chunk: 5  Samples Per Chunk: 1  Sample Description ID: 24

	WAVEFORMATEX *wfex;
	unsigned long ci;
	unsigned long co;
	unsigned long limit;

	bool ret = false;

	// Be aware of what has already been Verified (see above).

	if (default_sample_size == 0) goto Abort;

	wfex = (WAVEFORMATEX *)pFormat;

	if (wfex->nChannels != 1 && wfex->nChannels != 2) goto Abort;
	if (wfex->wBitsPerSample != 8 && wfex->wBitsPerSample != 16) goto Abort;

	if (dwFormat == A_IN24)
		wfex->wBitsPerSample = 24;
	else if (dwFormat == A_IN32)
		wfex->wBitsPerSample = 32;

	// Block alignment is needed for parsing the chunks.
	wfex->nBlockAlign = wfex->nChannels * (wfex->wBitsPerSample / 8);

	perceived_samples = 0;
	num_samples = num_chunk_offsets;

	sample_list = new QTSampleInfo[num_samples];
	if (sample_list == NULL) goto Abort;

	for (ci = 0; ci < num_samples; ++ci)
	{
		// Note: zero is not a valid position
		sample_list[ci].pos  = 0;
		sample_list[ci].size = 0;
	}

	for (ci = 0; ci < num_chunk_infos; ++ci)
	{
		if ((ci + 1) < num_chunk_infos)
		{
			limit = chunk_info[ci + 1].first;
		}
		else
		{
			limit = num_chunk_offsets + 1;
		}

		for (co = chunk_info[ci].first; co < limit; ++co)
		{
			if (co < 1 || co > num_chunk_offsets) goto Abort;

			if (chunk_offset32 == NULL)
			{
				sample_list[co - 1].pos = chunk_offset64[co - 1];
			}
			else
			{
				sample_list[co - 1].pos = chunk_offset32[co - 1];
			}

			sample_list[co - 1].size = chunk_info[ci].samples * wfex->nBlockAlign;

			perceived_samples += chunk_info[ci].samples;
		}
	}

	// Calculate total bytes and maximum sample size

	dwMaxSample = sample_list[0].size;
	i64TotalBytes = dwMaxSample;

	for (ci = 1; ci < num_samples; ++ci)
	{
		co = sample_list[ci].size;
		if (co > dwMaxSample) dwMaxSample = co;
		i64TotalBytes += co;
	}

	// 'sowt' is little-endian
	if (dwFormat == A_SOWT && wfex->wBitsPerSample >= 16)
	{
		iEndianness = 1;
	}

	wfex->wFormatTag      = WAVE_FORMAT_PCM;
	wfex->nAvgBytesPerSec = wfex->nBlockAlign * wfex->nSamplesPerSec;
	wfex->cbSize          = 0;
	
	dwRate  = wfex->nSamplesPerSec * wfex->nBlockAlign;
	dwScale = wfex->nBlockAlign;

	ret = true;		// All OK

Abort:
	return ret;
}


/***************************************************************
 *          Adaptive Multi-Rate Narrowband
 *          Adaptive Multi-Rate Wideband
 ***************************************************************/

bool QTTrack_samr::Finish()
{
	// Derive sample positions using the chunk tables

	WAVEFORMATEX *wfex;

	unsigned long ci;
	unsigned long co;
	unsigned long limit;

	unsigned long sample;
	unsigned long s;
	__int64 pos;

	bool ret = false;

	wfex = (WAVEFORMATEX *)pFormat;

	if (dwFormat == A_SAMR)
	{
		wfex->nSamplesPerSec = 8000;
		wfex->wFormatTag = WAVE_FORMAT_SAMR;
	}
	else	// AMR-WB
	{
		wfex->nSamplesPerSec = 16000;
		wfex->wFormatTag = WAVE_FORMAT_SAWB;
	}

	// Is it possible for AMR to be stereo?
	wfex->nChannels = 1;

	// Be aware of what has already been Verified (see above).

	perceived_samples = num_samples;
	sample_list = new QTSampleInfo[num_samples];
	if (sample_list == NULL) goto Abort;

	for (ci = 0; ci < num_samples; ++ci)
	{
		sample_list[ci].pos = 0;
	}

	// Derive sample sizes
	if (default_sample_size != 0)
	{
		for (ci = 0; ci < num_samples; ++ci)
		{
			sample_list[ci].size = default_sample_size;
		}
	}
	else
	{
		for (ci = 0; ci < num_samples; ++ci)
		{
			sample_list[ci].size = sample_size[ci];
		}
	}

	sample = 0;

	for (ci = 0; ci < num_chunk_infos; ++ci)
	{
		if ((ci + 1) < num_chunk_infos)
		{
			limit = chunk_info[ci + 1].first;
		}
		else
		{
			limit = num_chunk_offsets + 1;
		}

		for (co = chunk_info[ci].first; co < limit; ++co)
		{
			if (co < 1 || co > num_chunk_offsets) goto Abort;

			if (chunk_offset32 == NULL)
			{
				pos = chunk_offset64[co - 1];
			}
			else
			{
				pos = chunk_offset32[co - 1];
			}
				
			for (s = 0; s < chunk_info[ci].samples; ++s)
			{
				if (sample >= num_samples) goto Abort;
				sample_list[sample].pos = pos;
				pos += sample_list[sample].size;
				++sample;
			}
		}
	}

	// Calculate total bytes and maximum sample size

	dwMaxSample = sample_list[0].size;

	if (dwFlags & QTT_ADD_AMR)
		dwMaxSample += (dwFormat == A_SAMR)? 6: 9;

	i64TotalBytes = dwMaxSample;

	for (ci = 1; ci < num_samples; ++ci)
	{
		co = sample_list[ci].size;
		i64TotalBytes += co;

		if (co != dwMaxSample)
		{
			dwFlags |= QTT_HAS_VSB;
			if (co > dwMaxSample) dwMaxSample = co;
		}
	}

	// There are 160 samples per AMR-NB frame, frequency is 8000 Hz.
	// There are 320 samples per AMR-WB frame, frequency is 16000 Hz.
	// Thus there are always 50 frames per second.

	// Seen so far:
	// AMR-NB, mono 12.2k = 32 bytes per frame
	// AMR-WB, mono 12.2k = 61 bytes per frame (?)

	if (dwSamplesPerPacket == 0)
	{
		// Derive the number of samples per packet:
		// Duration_in_secs * frequency_in_Hz = Total_samples
		// Total_samples / Total_packets = Samples_per_packet
		// Remember your algebra: (a / b) / c = a / (b * c)

		dwSamplesPerPacket = (unsigned long)(
			((double)i64Duration * (double)wfex->nSamplesPerSec) /
			((double)num_samples * (double)dwTimeScale) + 0.5
		);

		// Round to an integral multiple of 160
		dwSamplesPerPacket = (dwSamplesPerPacket + 80) / 160;
		dwSamplesPerPacket *= 160;

		if (dwSamplesPerPacket == 0) goto Abort;
	}

	wfex->nAvgBytesPerSec = (DWORD)(
		((double)i64TotalBytes * (double)wfex->nSamplesPerSec) /
		((double)dwSamplesPerPacket * (double)num_samples) + 0.5
	);
	
	wfex->nBlockAlign    = (WORD)dwMaxSample;
	wfex->wBitsPerSample = 0;
	wfex->cbSize         = 0;

	dwRate  = wfex->nSamplesPerSec;
	dwScale = dwSamplesPerPacket;

	ret = true;		// All OK

Abort:
	return ret;
}


/***************************************************************
 *          A-law, mu-law (logarithmic PCM)
 ***************************************************************/

bool QTTrack_ulaw::Finish()
{
	// Derive sample positions using the chunk tables

	WAVEFORMATEX *wfex;
	unsigned long ci;
	unsigned long co;
	unsigned long limit;

	bool ret = false;

	// Be aware of what has already been Verified (see above).

	if (default_sample_size == 0) goto Abort;

	wfex = (WAVEFORMATEX *)pFormat;

	if (wfex->nChannels != 1 && wfex->nChannels != 2) goto Abort;
	if (wfex->wBitsPerSample != 16) goto Abort;

	// A-law and mu-law code one 16-bit sample in 8 bits
	wfex->wBitsPerSample = 8;

	// Block alignment is needed for parsing the chunks.
	wfex->nBlockAlign = wfex->nChannels * (wfex->wBitsPerSample / 8);

	perceived_samples = 0;
	num_samples = num_chunk_offsets;

	sample_list = new QTSampleInfo[num_samples];
	if (sample_list == NULL) goto Abort;

	for (ci = 0; ci < num_samples; ++ci)
	{
		// Note: zero is not a valid position
		sample_list[ci].pos  = 0;
		sample_list[ci].size = 0;
	}

	// samples per chunk flipflops between 11021 and 11022
	// last chunk has 3576 samples
	// default sample size is 1

	// total sample count is 2692779

	for (ci = 0; ci < num_chunk_infos; ++ci)
	{
		if ((ci + 1) < num_chunk_infos)
		{
			limit = chunk_info[ci + 1].first;
		}
		else
		{
			limit = num_chunk_offsets + 1;
		}

		for (co = chunk_info[ci].first; co < limit; ++co)
		{
			if (co < 1 || co > num_chunk_offsets) goto Abort;

			if (chunk_offset32 == NULL)
			{
				sample_list[co - 1].pos = chunk_offset64[co - 1];
			}
			else
			{
				sample_list[co - 1].pos = chunk_offset32[co - 1];
			}

			sample_list[co - 1].size = chunk_info[ci].samples * wfex->nBlockAlign;

			perceived_samples += chunk_info[ci].samples;
		}
	}

	// Calculate total bytes and maximum sample size

	dwMaxSample   = sample_list[0].size;
	i64TotalBytes = dwMaxSample;

	for (ci = 1; ci < num_samples; ++ci)
	{
		co = sample_list[ci].size;
		if (co > dwMaxSample) dwMaxSample = co;
		i64TotalBytes += co;
	}

	if (dwFormat == A_ULAW)
		wfex->wFormatTag  = WAVE_FORMAT_MULAW;
	else
		wfex->wFormatTag  = WAVE_FORMAT_ALAW;

	wfex->nAvgBytesPerSec = wfex->nBlockAlign * wfex->nSamplesPerSec;
	wfex->cbSize          = 0;

	dwRate  = wfex->nSamplesPerSec * wfex->nBlockAlign;
	dwScale = wfex->nBlockAlign;

	ret = true;		// All OK

Abort:
	return ret;
}


