///////////////////////////////////////////////////////////////////////
// QuickTime plugin for VirtualDub 1.8.8+
// Copyright (C) 2007-2012 fccHandler
// Copyright (C) 1998-2012 Avery Lee
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
///////////////////////////////////////////////////////////////////////

// Main Plugin interfaces

#if defined(_MSC_VER) && (_MSC_VER >= 1400)
	#define _CRT_SECURE_NO_WARNINGS
#endif

#define WIN32_LEAN_AND_MEAN
#include <windows.h>
#include <commctrl.h>
#include <vfw.h>
#include <stdio.h>
#include <crtdbg.h>

#include "vd2\plugin\vdinputdriver.h"
#include "Unknown.h"

#include "QTTypes.h"
#include "QTTrack.h"

#include "WinFile.h"
#include "ErrorLog.h"
#include "InputOptionsQT.h"
#include "InputFileQT.h"
#include "VideoSourceQT.h"
#include "VideoCodec.h"


///////////////////////////////////////////////////////////////////////
//
//						VideoDecoderQT
//
///////////////////////////////////////////////////////////////////////

class VideoDecoderQT : public vdxunknown<IVDXVideoDecoder> {

private:
	InputFileQT      *const parentPtr;
	QTTrack          *const track;

	VideoCodec       mCodec;

	sint64           mPrevFrame;
	DWORD            dwActualFrame;

public:
	VideoDecoderQT(InputFileQT *const pp, QTTrack *const qtt);
	~VideoDecoderQT();

	// IVDXVideoDecoder methods
	const void *     VDXAPIENTRY DecodeFrame(const void *inputBuffer, uint32 data_len, bool is_preroll, sint64 sampleNumber, sint64 targetFrame);
	uint32           VDXAPIENTRY GetDecodePadding();
	void             VDXAPIENTRY Reset();
	bool             VDXAPIENTRY IsFrameBufferValid();
	const VDXPixmap& VDXAPIENTRY GetFrameBuffer();
	bool             VDXAPIENTRY SetTargetFormat(int format, bool useDIBAlignment);
	bool             VDXAPIENTRY SetDecompressedFormat(const VDXBITMAPINFOHEADER *pbih);
	bool             VDXAPIENTRY IsDecodable(sint64 sample_num);
	const void *     VDXAPIENTRY GetFrameBufferBase();
};


VideoDecoderQT::VideoDecoderQT(InputFileQT *const pp, QTTrack *const qtt)
	: parentPtr(pp)
	, track(qtt)
{
	parentPtr->AddRef();
	parentPtr->vCodec = NULL;

	if (track->fccType == streamtypeVIDEO)
	{
		dwActualFrame = 0xFFFFFFFF;
		mPrevFrame = -1;
	}
	else
	{
		dwActualFrame = 0;
		mPrevFrame = 0;
	}
}


VideoDecoderQT::~VideoDecoderQT()
{
	parentPtr->vCodec = NULL;
	parentPtr->Release();
}

#ifdef _DEBUG
static void Dump(const wchar_t *pFile, const void *pData, int bytes)
{
	if (pFile != NULL && pData != NULL && bytes > 0)
	{
		HANDLE h = CreateFileW(pFile, GENERIC_WRITE, FILE_SHARE_READ,
			NULL, OPEN_ALWAYS, FILE_ATTRIBUTE_NORMAL, NULL);
		if (h != INVALID_HANDLE_VALUE)
		{
			DWORD dwWritten;
			WriteFile(h, pData, bytes, &dwWritten, NULL);
			SetEndOfFile(h);
			CloseHandle(h);
		}
	}
}
#endif

const void *VideoDecoderQT::DecodeFrame(
	const void *inputBuffer, uint32 data_len, bool is_preroll,
	sint64 streamFrame,	sint64 targetFrame)
{

#if defined(_MSC_VER) && (_MSC_VER >= 1400)
_RPT5(_CRT_WARN, "VideoDecoderQT::DecodeFrame(0x%p, %lu, %s, %I64i, %I64i)\n",
	inputBuffer, data_len, is_preroll? "true": "false", streamFrame, targetFrame);
#endif

	if (track->fccType == streamtypeVIDEO)
	{
		if (streamFrame < 0) streamFrame = targetFrame;
		
		if (streamFrame >= 0 && streamFrame < (sint64)track->perceived_samples)
		{
			DWORD dwFrame = (DWORD)streamFrame;

			dwFrame = track->sample_map[dwFrame];

			// If dwFrame is the same as dwActualFrame,
			// then we've already decoded this frame.

			if (dwFrame != dwActualFrame)
			{
				DWORD f;

				if (track->sample_list[dwFrame].size & QT_KEYFRAME) {
					f = 0;
				} else {
					f = ICDECOMPRESS_NOTKEYFRAME;
				}

				if (is_preroll)
				{
					f |= ICDECOMPRESS_PREROLL;
					dwActualFrame = 0xFFFFFFFF;
					mPrevFrame = -1;
				}
				else
				{
					dwActualFrame = dwFrame;
					mPrevFrame = targetFrame;

					// This helps with some manifestations of the
					// ffdshow-tryouts green frame problem, but not all.
					//if (dwFrame == 0) mCodec.ClearBuffer();
				}

				mCodec.Decompress((LPVOID)inputBuffer, data_len, f);
			}
		}
	}

	return mCodec.mFrameBuffer;
}


uint32 VideoDecoderQT::GetDecodePadding()
{
	return 0;
}


void VideoDecoderQT::Reset()
{
_RPT0(_CRT_WARN, "VideoDecoderQT::Reset()\n");
	if (track->fccType == streamtypeVIDEO)
	{
		mPrevFrame = -1;
		dwActualFrame = 0xFFFFFFFF;
	}
}


bool VideoDecoderQT::IsFrameBufferValid()
{
_RPT1(_CRT_WARN, "VideoDecoderQT::IsFrameBufferValid() returning %s\n", (mPrevFrame >= 0)? "true": "false");
	return (mPrevFrame >= 0);
}


const VDXPixmap& VideoDecoderQT::GetFrameBuffer()
{
_RPT0(_CRT_WARN, "VideoDecoderQT::GetFrameBuffer()\n");
	return mCodec.mPixmap;
}


bool VideoDecoderQT::SetTargetFormat(int format, bool useDIBAlignment)
{
_RPT2(_CRT_WARN, "VideoDecoderQT::SetTargetFormat(%i, %s)\n",
	format, (useDIBAlignment)? "true": "false");

	if (track->fccType == streamtypeVIDEO)
	{
		Reset();

		if (mCodec.Init((BITMAPINFOHEADER *)track->pFormat, format))
		{
			mCodec.ClearBuffer();
			parentPtr->vCodec = mCodec.hCodec;
			return true;
		}

		if (mCodec.hCodec != NULL) return false;
		parentPtr->vCodec = NULL;

		// Missing codec
		{
			CHAR inform[32];
			unsigned char fcc[5];
			int i;

			*(unsigned long *)fcc = ((BITMAPINFOHEADER *)track->pFormat)->biCompression;
			for (i = 0; i < 4; ++i)
			{
				if (fcc[i] < 32) fcc[i] = ' ';
			}
			fcc[i] = '\0';

			_snprintf(inform, sizeof(inform), "MISSING CODEC (%s)", (char *)fcc);
			inform[sizeof(inform) - 1] = '\0';

			mCodec.Inform(inform);
		}

		return (format == mCodec.mPixmap.format ||
			format == nsVDXPixmap::kPixFormat_Null);
	}

	// Emulated video source

	if (mCodec.Init(NULL, format))
	{
		// No video stream
		mCodec.Inform("NO VIDEO");

		return (format == mCodec.mPixmap.format
		     || format == nsVDXPixmap::kPixFormat_Null);
	}

	return false;
}


bool VideoDecoderQT::SetDecompressedFormat(const VDXBITMAPINFOHEADER *pbih)
{
	return false;
}


const void *VideoDecoderQT::GetFrameBufferBase()
{
_RPT0(_CRT_WARN, "VideoDecoderQT::GetFrameBufferBase()\n");

	return mCodec.mFrameBuffer;
}


bool VideoDecoderQT::IsDecodable(sint64 sample_num64)
{
_RPT1(_CRT_WARN, "VideoDecoderQT::IsDecodable(%I64)\n", sample_num64);

	if (track->fccType == streamtypeVIDEO && sample_num64 != (mPrevFrame + 1))
	{
		DWORD dwSample = (DWORD)sample_num64;

		if (dwSample < track->perceived_samples)
		{
			dwSample = track->sample_map[dwSample];
			return ((track->sample_list[dwSample].size & QT_KEYFRAME) != 0);
		}
		return false;
	}
	return true;
}


///////////////////////////////////////////////////////////////////////
//
//                      VideoSourceQT
//
///////////////////////////////////////////////////////////////////////


VideoSourceQT::VideoSourceQT(InputFileQT *const pp, QTTrack *const qtt)
	: parentPtr(pp)
	, video(qtt)
{
	parentPtr->AddRef();
	errmode = kErrorModeDecodeAnyway;

	dwPrevSample = 0xFFFFFFFF;

	unsigned long max = (video->dwMaxSample < 4)? 4: video->dwMaxSample;
	frame_buffer = new unsigned char[max];
}


VideoSourceQT::~VideoSourceQT()
{
	delete[] frame_buffer;
	parentPtr->Release();
}


int VideoSourceQT::AddRef()
{
	return vdxunknown<IVDXStreamSource>::AddRef();
}


int VideoSourceQT::Release()
{
	return vdxunknown<IVDXStreamSource>::Release();
}


void *VideoSourceQT::AsInterface(uint32 iid)
{
	if (iid == IVDXVideoSource::kIID)
		return static_cast<IVDXVideoSource *>(this);

	if (iid == IVDXStreamSourceV3::kIID)
		return static_cast<IVDXStreamSourceV3 *>(this);

	return vdxunknown<IVDXStreamSource>::AsInterface(iid);
}


bool VideoSourceQT::IsDroppedFrame(unsigned long frame)
{
	if (frame != 0)
	{
		return (video->sample_map[frame] == video->sample_map[frame - 1]);
	}
	return false;
}


void VideoSourceQT::GetStreamSourceInfo(VDXStreamSourceInfo& info)
{
	info.mSampleRate.mNumerator     = video->dwRate;
	info.mSampleRate.mDenominator   = video->dwScale;
	info.mSampleCount               = video->perceived_samples;

	if (video->hSpacing > 0 && video->vSpacing > 0)
	{
		info.mPixelAspectRatio.mNumerator   = video->hSpacing;
		info.mPixelAspectRatio.mDenominator = video->vSpacing;
	}
	else
	{
		info.mPixelAspectRatio.mNumerator   = 1;
		info.mPixelAspectRatio.mDenominator = 1;
	}
}


void VideoSourceQT::GetStreamSourceInfoV3(VDXStreamSourceInfoV3& info)
{
	const BITMAPINFOHEADER *bmih = (BITMAPINFOHEADER *)video->pFormat;

	GetStreamSourceInfo(info.mInfo);

	if (video->dwFlags & QTT_HAS_VSB)
		info.mFlags = VDXStreamSourceInfoV3::kFlagVariableSizeSamples;
	else
		info.mFlags = 0;

	info.mfccHandler = parentPtr->UppercaseFormat(bmih->biCompression);
}


int VideoSourceQT::ParseAVCFrame(const unsigned char *src, unsigned char *dst, int len)
{
	// Unwrap NALUs and return the total size
	// Use special-case code to improve speed (hopefully)

	int ret = 0;

	__try {
#if 1
		switch (video->avc_len)
		{

		case 1:
			if (dst == NULL)
			{
				while (len > 1)
				{
					int size = *src++;
					--len;
					if (size > len) size = len;
					ret += size + 4;
					src += size;
					len -= size;
				}
			}
			else
			{
				while (len > 1)
				{
					int size = *src++;
					--len;
					if (size > len) size = len;
					*(unsigned long *)dst = 0x01000000;
					memcpy(dst + 4, src, size);
					dst += size + 4;
					ret += size + 4;
					src += size;
					len -= size;
				}
			}
			break;

		case 2:
			if (dst == NULL)
			{
				while (len > 2)
				{
					int size = *src++;
					size = (size << 8) + *src++;
					len -= 2;
					if (size > len) size = len;
					ret += size + 4;
					src += size;
					len -= size;
				}
			}
			else
			{
				while (len > 2)
				{
					int size = *src++;
					size = (size << 8) + *src++;
					len -= 2;
					if (size > len) size = len;
					*(unsigned long *)dst = 0x01000000;
					memcpy(dst + 4, src, size);
					dst += size + 4;
					ret += size + 4;
					src += size;
					len -= size;
				}
			}
			break;

		case 3:
			if (dst == NULL)
			{
				while (len > 3)
				{
					int size = *src++;
					size = (size << 8) + *src++;
					size = (size << 8) + *src++;
					len -= 3;
					if (size > len) size = len;
					ret += size + 4;
					src += size;
					len -= size;
				}
			}
			else
			{
				while (len > 3)
				{
					int size = *src++;
					size = (size << 8) + *src++;
					size = (size << 8) + *src++;
					len -= 3;
					if (size > len) size = len;
					*(unsigned long *)dst = 0x01000000;
					memcpy(dst + 4, src, size);
					dst += size + 4;
					ret += size + 4;
					src += size;
					len -= size;
				}
			}
			break;

		case 4:
			if (dst == NULL)
			{
				while (len > 4)
				{
					int size = *src++;
					size = (size << 8) + *src++;
					size = (size << 8) + *src++;
					size = (size << 8) + *src++;
					len -= 4;
					if (size > len) size = len;
					ret += size + 4;
					src += size;
					len -= size;
				}
			}
			else
			{
				while (len > 4)
				{
					int size = *src++;
					size = (size << 8) + *src++;
					size = (size << 8) + *src++;
					size = (size << 8) + *src++;
					len -= 4;
					if (size > len) size = len;
					*(unsigned long *)dst = 0x01000000;
					memcpy(dst + 4, src, size);
					dst += size + 4;
					ret += size + 4;
					src += size;
					len -= size;
				}
			}
			break;
		}
#else
		while (len > video->avc_len)
		{
			int size = *src++;
			--len;
			if (video->avc_len > 1)
			{
				size = (size << 8) + *src++;
				--len;
				if (video->avc_len > 2)
				{
					size = (size << 8) + *src++;
					size = (size << 8) + *src++;
					len -= 2;
				}
			}

			if (size > len) size = len;

			if (dst != NULL)
			{
				*dst++ = 0;
				*dst++ = 0;
				*dst++ = 0;
				*dst++ = 1;
				memcpy(dst, src, size);
				dst += size;
			}

			ret += size + 4;
			src += size;
			len -= size;
		}
#endif
	}
	__except(EXCEPTION_EXECUTE_HANDLER)
	{
		ret = 0;
	}

	return ret;
}


bool VideoSourceQT::Read(sint64 lStart64, uint32 lCount, void *lpBuffer, uint32 cbBuffer, uint32 *lBytesRead, uint32 *lSamplesRead)
{
	unsigned char *dst = (unsigned char *)lpBuffer;
	int framesize;
	int actual;

	DWORD dwStart = (DWORD)lStart64;

//_RPT2(_CRT_WARN, "VideoSourceQT::Read %lu, %lu\n", dwStart, dwPrevSample);

	if (frame_buffer == NULL || dwStart >= video->perceived_samples)
	{
		if (lBytesRead) *lBytesRead = 0;
		if (lSamplesRead) *lSamplesRead = 0;
		return true;
	}

	// Dropped frame?
	if (IsDroppedFrame(dwStart))
	{
		if (lBytesRead) *lBytesRead = 0;
		if (lSamplesRead) *lSamplesRead = 1;
		return true;
	}
	dwStart = video->sample_map[dwStart];

	const QTSampleInfo& si = video->sample_list[dwStart];

	framesize = (int)(si.size & (~QT_KEYFRAME));

	if (video->dwFormat == V_AVC1)
	{
		// read data, but only if it's new data
		if (dwStart != dwPrevSample)
		{
			dwPrevSample = dwStart;
			parentPtr->ReadWholeSample(frame_buffer, si);
		}
		
		actual = ParseAVCFrame(frame_buffer, NULL, framesize);
		if (actual == 0)
		{
			// Workaround for apparently corrupt file:
			// E-Clip0381  Kathryn Beaumont - On an Island With You 1948.mp4

			if (lBytesRead) *lBytesRead = 0;
			if (lSamplesRead) *lSamplesRead = 0;
			return true;
		}
	}
	else
	{
		actual = framesize;
	}
	if (si.size & QT_KEYFRAME) actual += video->dcr_size;

	// Size query?
	if (dst == NULL || cbBuffer < (unsigned int)actual)
	{
		if (lBytesRead) *lBytesRead = actual;
		if (lSamplesRead) *lSamplesRead = 1;
		return (dst == NULL);
	}

	if (video->dwFormat == V_AVC1)
	{
		if (si.size & QT_KEYFRAME)
		{
			memcpy(dst, video->avc_dcr, video->dcr_size);
			dst += video->dcr_size;
		}
		dst += ParseAVCFrame(frame_buffer, dst, framesize);
	}
	else
	{
		// read data, but only if it's new data
		if (dwStart != dwPrevSample)
		{
			dwPrevSample = dwStart;
			parentPtr->ReadWholeSample(frame_buffer, si);
		}
		if (video->dwFormat == V_MP4V && (si.size & QT_KEYFRAME))
		{
			memcpy(dst, video->avc_dcr, video->dcr_size);
			dst += video->dcr_size;
		}
		memcpy(dst, frame_buffer, framesize);
		dst += framesize;
	}

	if (lBytesRead) *lBytesRead = (uint32)(dst - (unsigned char *)lpBuffer);
	if (lSamplesRead) *lSamplesRead = 1;
	return true;
}


const void *VideoSourceQT::GetDirectFormat()
{
	return video->pFormat;
}


int VideoSourceQT::GetDirectFormatLen()
{
	return video->lFormatLen;
}


IVDXStreamSource::ErrorMode VideoSourceQT::GetDecodeErrorMode()
{
	return errmode;
}


void VideoSourceQT::SetDecodeErrorMode(IVDXStreamSource::ErrorMode mode)
{
	errmode = mode;
}


bool VideoSourceQT::IsDecodeErrorModeSupported(IVDXStreamSource::ErrorMode mode)
{
	return true;
}


bool VideoSourceQT::IsVBR()
{
	return false;
}


sint64 VideoSourceQT::TimeToPositionVBR(sint64 us)
{
	return 0;
}


sint64 VideoSourceQT::PositionToTimeVBR(sint64 samples)
{
	return 0;
}


void VideoSourceQT::GetVideoSourceInfo(VDXVideoSourceInfo& info)
{
	const BITMAPINFOHEADER *bmih = (BITMAPINFOHEADER *)video->pFormat;

	info.mWidth        = bmih->biWidth;
	info.mHeight       = bmih->biHeight;
	info.mFlags        = 0;
	info.mDecoderModel = VDXVideoSourceInfo::kDecoderModelDefaultIP;
}


bool VideoSourceQT::CreateVideoDecoderModel(IVDXVideoDecoderModel **ppModel)
{
	*ppModel = NULL;
	return false;
}


bool VideoSourceQT::CreateVideoDecoder(IVDXVideoDecoder **ppDecoder)
{
	if (!parentPtr->pOptions->DeferDecoding())
	{
		*ppDecoder = new VideoDecoderQT(parentPtr, video);
		if (*ppDecoder != NULL)
		{
			(*ppDecoder)->AddRef();
			return true;
		}
	}
	*ppDecoder = NULL;
	return false;
}


void VideoSourceQT::GetSampleInfo(sint64 sample_num, VDXVideoFrameInfo& frameInfo)
{
	unsigned long i = (unsigned long)sample_num;
	if (i < video->perceived_samples && !IsDroppedFrame(i))
	{
		i = video->sample_map[i];

		const QTSampleInfo& si = video->sample_list[i];

		if (si.size & QT_KEYFRAME)
		{
			frameInfo.mTypeChar  = 'K';
			frameInfo.mFrameType = kVDXVFT_Independent;
		}
		else
		{
			frameInfo.mTypeChar  = ' ';
			frameInfo.mFrameType = kVDXVFT_Predicted;
		}

		frameInfo.mBytePosition  = si.pos;
		return;
	}

	// Bad query or dropped frame
	frameInfo.mTypeChar     = 'D';
	frameInfo.mFrameType    = kVDXVFT_Null;
	frameInfo.mBytePosition = -1;
}


bool VideoSourceQT::IsKey(sint64 sample)
{
	unsigned long i = (unsigned long)sample;
	if (i < video->perceived_samples && !IsDroppedFrame(i))
	{
		i = video->sample_map[i];
		return ((video->sample_list[i].size & QT_KEYFRAME) != 0);
	}
	return false;
}


sint64 VideoSourceQT::GetFrameNumberForSample(sint64 sample_num)
{
	return sample_num;
}


sint64 VideoSourceQT::GetSampleNumberForFrame(sint64 display_num)
{
	return display_num;
}


sint64 VideoSourceQT::GetRealFrame(sint64 display_num)
{
	return display_num;
}


sint64 VideoSourceQT::GetSampleBytePosition(sint64 sample_num)
{
	unsigned long i = (unsigned long)sample_num;
	if (i < video->perceived_samples && !IsDroppedFrame(i))
	{
		i = video->sample_map[i];
		return video->sample_list[i].pos;
	}
	return -1;
}


///////////////////////////////////////////////////////////////////////
//
//                      VideoSourceDummy
//
///////////////////////////////////////////////////////////////////////


VideoSourceDummy::VideoSourceDummy(InputFileQT *const pp, QTTrack *const qtt)
	: parentPtr(pp)
	, audio(qtt)
{
	parentPtr->AddRef();
	errmode = kErrorModeDecodeAnyway;

	// Dummy video, Direct Stream Copy format.
	// This must match the VideoDecoderQT dummy format!

	bmDummy.biSize          = sizeof(bmDummy);
	bmDummy.biWidth         = 320;
	bmDummy.biHeight        = 240;
	bmDummy.biPlanes        = 1;
	bmDummy.biBitCount      = 32;
	bmDummy.biCompression   = BI_RGB;
	bmDummy.biSizeImage     = (bmDummy.biWidth * 4) * bmDummy.biHeight;
	bmDummy.biXPelsPerMeter = 0;
	bmDummy.biYPelsPerMeter = 0;
	bmDummy.biClrUsed       = 0;
	bmDummy.biClrImportant  = 0;
}


VideoSourceDummy::~VideoSourceDummy()
{
	parentPtr->Release();
}


int VideoSourceDummy::AddRef()
{
	return vdxunknown<IVDXStreamSource>::AddRef();
}


int VideoSourceDummy::Release()
{
	return vdxunknown<IVDXStreamSource>::Release();
}


void *VideoSourceDummy::AsInterface(uint32 iid)
{
	if (iid == IVDXVideoSource::kIID)
		return static_cast<IVDXVideoSource *>(this);

	if (iid == IVDXStreamSourceV3::kIID)
		return static_cast<IVDXStreamSourceV3 *>(this);

	return vdxunknown<IVDXStreamSource>::AsInterface(iid);
}


void VideoSourceDummy::GetStreamSourceInfo(VDXStreamSourceInfo& info)
{
	const WAVEFORMATEX *wfex = (WAVEFORMATEX *)audio->pFormat;

	info.mSampleRate.mNumerator   = wfex->nSamplesPerSec;
	info.mSampleRate.mDenominator = 1000;

	info.mSampleCount = (sint64)(
	    ((double)info.mSampleRate.mNumerator   * (double)audio->i64Duration)
	  / ((double)info.mSampleRate.mDenominator * (double)audio->dwTimeScale)
	);

	// Err on the side of caution to avoid cutting off any audio.
	++info.mSampleCount;

	// TODO:
	info.mPixelAspectRatio.mNumerator   = 1;
	info.mPixelAspectRatio.mDenominator = 1;
}


void VideoSourceDummy::GetStreamSourceInfoV3(VDXStreamSourceInfoV3& info)
{
	GetStreamSourceInfo(info.mInfo);

	info.mFlags      = 0;
	info.mfccHandler = 0;
}


bool VideoSourceDummy::Read(sint64 lStart64, uint32 lCount, void *lpBuffer, uint32 cbBuffer, uint32 *lBytesRead, uint32 *lSamplesRead)
{
	// Buffer size query?
	if (lpBuffer == NULL)
	{
		if (lBytesRead) *lBytesRead = bmDummy.biSizeImage;
		if (lSamplesRead) *lSamplesRead = 1;
		return true;
	}

	// Buffer too small?
	if (cbBuffer < bmDummy.biSizeImage)
	{
		if (lBytesRead) *lBytesRead = 0;
		if (lSamplesRead) *lSamplesRead = 0;
		return false;
	}

	if (lStart64 == 0i64)
	{
		// Gad this is such a hack...
		VideoDecoderQT vd(parentPtr, audio);
		vd.SetTargetFormat(nsVDXPixmap::kPixFormat_Null, true);
		memcpy(lpBuffer, vd.GetFrameBufferBase(), bmDummy.biSizeImage);

		if (lBytesRead) *lBytesRead = bmDummy.biSizeImage;
	}
	else
	{
		if (lBytesRead) *lBytesRead = 0;
	}

	if (lSamplesRead) *lSamplesRead = 1;

	return true;
}


const void *VideoSourceDummy::GetDirectFormat()
{
	return &bmDummy;
}


int VideoSourceDummy::GetDirectFormatLen()
{
	return (int)sizeof(bmDummy);
}


IVDXStreamSource::ErrorMode VideoSourceDummy::GetDecodeErrorMode()
{
	return errmode;
}


void VideoSourceDummy::SetDecodeErrorMode(IVDXStreamSource::ErrorMode mode)
{
	errmode = mode;
}


bool VideoSourceDummy::IsDecodeErrorModeSupported(IVDXStreamSource::ErrorMode mode)
{
	return true;
}


bool VideoSourceDummy::IsVBR()
{
	return false;
}


sint64 VideoSourceDummy::TimeToPositionVBR(sint64 us)
{
	return 0;
}


sint64 VideoSourceDummy::PositionToTimeVBR(sint64 samples)
{
	return 0;
}


void VideoSourceDummy::GetVideoSourceInfo(VDXVideoSourceInfo& info)
{
	info.mWidth        = bmDummy.biWidth;
	info.mHeight       = bmDummy.biHeight;
	info.mFlags        = 0;
	info.mDecoderModel = VDXVideoSourceInfo::kDecoderModelDefaultIP;
}


bool VideoSourceDummy::CreateVideoDecoderModel(IVDXVideoDecoderModel **ppModel)
{
	*ppModel = NULL;
	return false;
}


bool VideoSourceDummy::CreateVideoDecoder(IVDXVideoDecoder **ppDecoder)
{
	*ppDecoder = NULL;
	return false;
}


void VideoSourceDummy::GetSampleInfo(sint64 sample_num, VDXVideoFrameInfo& frameInfo)
{
	if (sample_num == 0i64)
	{
		frameInfo.mTypeChar     = 'K';
		frameInfo.mFrameType    = kVDXVFT_Independent;
		frameInfo.mBytePosition = 0;
	}
	else
	{
		frameInfo.mTypeChar     = 'D';
		frameInfo.mFrameType    = kVDXVFT_Null;
		frameInfo.mBytePosition = -1;
	}
}


bool VideoSourceDummy::IsKey(sint64 sample)
{
	return (sample == 0i64);
}


sint64 VideoSourceDummy::GetFrameNumberForSample(sint64 sample_num)
{
	return sample_num;
}


sint64 VideoSourceDummy::GetSampleNumberForFrame(sint64 display_num)
{
	return display_num;
}


sint64 VideoSourceDummy::GetRealFrame(sint64 display_num)
{
	return display_num;
}


sint64 VideoSourceDummy::GetSampleBytePosition(sint64 sample_num)
{
	return -1;
}


