///////////////////////////////////////////////////////////////////////
// QuickTime plugin for VirtualDub 1.8.8+
// Copyright (C) 2007-2012 fccHandler
// Copyright (C) 1998-2012 Avery Lee
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
///////////////////////////////////////////////////////////////////////

// Windows file input functions, wrapped in a class
// Files larger than 4GB are supported on NTFS.


#define WIN32_LEAN_AND_MEAN
#include <windows.h>

#include "WinFile.h"


WinFile::WinFile()
{
	m_hFile         = INVALID_HANDLE_VALUE;
	m_Pos.QuadPart  = 0;
	m_Size.QuadPart = 0;

	memset(m_Buffer, 0, WF_BUFFERSIZE);
}


WinFile::~WinFile()
{
	if (m_hFile != INVALID_HANDLE_VALUE)
	{
		CloseHandle(m_hFile);
	}
}


/*
	How Windows handles some aberrant situations:

	Setting the file pointer to absolute -10 returns 0xFFFFFFFF
	and error code 87 (ERROR_INVALID_PARAMETER), and does
	not move the file pointer (apparently)

	Setting the file pointer to relative -10 returns 0xFFFFFFFF
	and error code 131 (ERROR_NEGATIVE_SEEK), and does
	not move the file pointer

	Setting the file pointer AT or BEYOND the end of file succeeds.

	Attempting to read past the end of file does not cause
	an error, but reads zero bytes.
*/

void WinFile::Refresh()
{
	DWORD dwHi = m_Pos.HighPart;
	DWORD dwLo = m_Pos.LowPart & -WF_BUFFERSIZE;

	SetFilePointer(m_hFile, dwLo, (PLONG)&dwHi, FILE_BEGIN);

	if (ReadFile(m_hFile, m_Buffer, WF_BUFFERSIZE, &dwLo, NULL) == 0)
	{
		dwLo = 0;
	}

	if (dwLo < WF_BUFFERSIZE)
	{
		memset(m_Buffer + dwLo, 0, WF_BUFFERSIZE - dwLo);
	}
}


// static
char *WinFile::WcharToAnsi(const wchar_t *lpszWchar)
{
	// Convert a wide character string to ANSI.
	if (lpszWchar != NULL)
	{
		int len = WideCharToMultiByte(CP_ACP, 0, lpszWchar, -1, NULL, 0, NULL, NULL);
		if (len > 1)
		{
			char *ret = new char[len + 2];
			if (ret != NULL)
			{
				if (WideCharToMultiByte(CP_ACP, 0, lpszWchar, -1, ret, len + 1, NULL, NULL) > 0)
				{
					ret[len + 1] = '\0';
					return ret;
				}
				delete[] ret;
			}
		}
	}
	return NULL;
}


bool WinFile::Open(HANDLE h)
{
	if (h != INVALID_HANDLE_VALUE)
	{
		DWORD dwHi = 0;
		DWORD dwLo = GetFileSize(h, &dwHi);

		if (dwLo != INVALID_FILE_SIZE || GetLastError() == NO_ERROR)
		{
			// file too big?
			if (dwHi < 0x80000000U)
			{
				if (m_hFile != INVALID_HANDLE_VALUE)
				{
					CloseHandle(m_hFile);
				}

				m_hFile         = h;
				m_Pos.QuadPart  = 0;
				m_Size.LowPart  = dwLo;
				m_Size.HighPart = dwHi;

				Refresh();
				return true;
			}
		}
		
		CloseHandle(h);
	}
	
	return false;
}


bool WinFile::OpenA(const char *lpFileName)
{
	HANDLE h = CreateFileA(
		lpFileName,
		GENERIC_READ,
		FILE_SHARE_READ,
		NULL,
		OPEN_EXISTING,
		FILE_FLAG_SEQUENTIAL_SCAN,
		NULL
	);

	return Open(h);
}


bool WinFile::OpenW(const wchar_t *lpFileName)
{
	HANDLE h = INVALID_HANDLE_VALUE;

	if (GetVersion() & 0x80000000U)
	{
		// Windows 95/98
		char *lpFileNameA = WcharToAnsi(lpFileName);
		if (lpFileNameA != NULL)
		{
			h = CreateFileA(
				lpFileNameA,
				GENERIC_READ,
				FILE_SHARE_READ,
				NULL,
				OPEN_EXISTING,
				FILE_FLAG_SEQUENTIAL_SCAN,
				NULL
			);

			delete[] lpFileNameA;
		}
	}
	else
	{
		// Windows NT/XP
		h = CreateFileW(
			lpFileName,
			GENERIC_READ,
			FILE_SHARE_READ,
			NULL,
			OPEN_EXISTING,
			FILE_FLAG_SEQUENTIAL_SCAN,
			NULL
		);
	}
	
	return Open(h);
}


bool WinFile::Seek(__int64 iAbsPos)
{
	if (m_hFile != INVALID_HANDLE_VALUE && iAbsPos >= 0)
	{
		LARGE_INTEGER OldPos;
		bool ret = true;

		// Do not seek past EOF
		if (iAbsPos > m_Size.QuadPart)
		{
			iAbsPos = m_Size.QuadPart;
			ret = false;
		}

		OldPos.QuadPart = m_Pos.QuadPart & -WF_BUFFERSIZE;
		m_Pos.QuadPart  = iAbsPos;

		if ((m_Pos.QuadPart & -WF_BUFFERSIZE) != OldPos.QuadPart)
		{
			Refresh();
		}

		return ret;
	}

	return false;
}


int WinFile::Read(void *lpBuffer, int nBytes)
{
	int ret = 0;

	if (m_Pos.QuadPart < m_Size.QuadPart
		&& lpBuffer != NULL
		&& nBytes > 0)
	{
		int remains, offset;

		// Don't read past EOF
		if ((__int64)nBytes > (m_Size.QuadPart - m_Pos.QuadPart))
		{
			nBytes = (int)(m_Size.QuadPart - m_Pos.QuadPart);
		}

		offset  = m_Pos.LowPart & (WF_BUFFERSIZE - 1);
		remains = WF_BUFFERSIZE - offset;

		for (;;)
		{
			if (remains > nBytes) remains = nBytes;
			memcpy(lpBuffer, m_Buffer + offset, remains);

			ret += remains;
			m_Pos.QuadPart += remains;

			nBytes -= remains;
			if (nBytes <= 0) break;

			lpBuffer = (char *)lpBuffer + remains;
			Refresh();

			offset  = 0;
			remains = WF_BUFFERSIZE;
		}

		// Finished exactly on a page boundary?
		if ((m_Pos.LowPart & (WF_BUFFERSIZE - 1)) == 0)
		{
			Refresh();
		}
	}

	return ret;
}


int WinFile::NextByte()
{
	int ret = 0;
	
	if (m_Pos.QuadPart < m_Size.QuadPart)
	{
		int offset = m_Pos.LowPart & (WF_BUFFERSIZE - 1);

		ret = m_Buffer[offset];
		++m_Pos.QuadPart;

		if (offset == (WF_BUFFERSIZE - 1)) Refresh();
	}

	return ret;
}

